# Import specific functions/classes from the common_utils module
from .common_utils import (
    create_context,
    extract_datamodel_from_raw_url,
    extract_subject_from_raw_url,
    generate_random_string,
    is_metadata_properly_reported,
    is_metadata_existed,
    is_url_existed,
    message_after_check_schema,
    normalized2keyvalues_v2,
    open_jsonref,
    open_yaml,
    parse_payload_v2,
    parse_property2ngsild_example,
    parse_yamlDict,
    schema_output_sum
)

# Define what gets imported when using `from pysmartdatamodels.utils import *`
__all__ = [
    create_context,
    extract_datamodel_from_raw_url,
    extract_subject_from_raw_url,
    generate_random_string,
    is_metadata_properly_reported,
    is_metadata_existed,
    is_url_existed,
    message_after_check_schema,
    normalized2keyvalues_v2,
    open_jsonref,
    open_yaml,
    parse_payload_v2,
    parse_property2ngsild_example,
    parse_yamlDict,
    schema_output_sum
]



# Import specific functions/classes from the utils module
from pysmartdatamodels.utils.common_utils import (
    create_context,
    extract_datamodel_from_raw_url,
    extract_subject_from_raw_url,
    generate_random_string,
    is_metadata_properly_reported,
    is_metadata_existed,
    is_url_existed,
    message_after_check_schema,
    normalized2keyvalues_v2,
    open_jsonref,
    open_yaml,
    parse_payload_v2,
    parse_property2ngsild_example,
    parse_yamlDict,
    schema_output_sum
)

# Define what gets imported when using `from pysmartdatamodels import *`
__all__ = [
    'attributes_datamodel',
    'datamodel_repolink',
    'datamodels_subject',
    'datatype_attribute',
    'description_attribute',
    'generate_sql_schema',
    'geojson_features_example_generator',
    'load_all_attributes',
    'load_all_datamodels',
    'list_all_subjects',
    'list_datamodel_metadata',
    'look_for_datamodel',
    'model_attribute',
    'ngsi_datatype_attribute',
    'ngsi_ld_example_generator',
    'ngsi_ld_keyvalue_example_generator',
    'print_datamodel',
    'subject_for_datamodel',
    'subject_repolink',
    'units_attribute',
    'update_broker',
    'update_data',
    'validate_data_model_schema',
    'validate_dcat_ap_distribution_sdm'
]


################################################################################
#  Licensed to the FIWARE Foundation (FF) under one
#  or more contributor license agreements. The FF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#  for questions address it to alberto.abella@fiware.org
################################################################################
import os
import json
import random
import re

import string
import jsonref
from typing import Dict, Tuple

from validator_collection import checkers


from faker import Faker
import requests
import rstr
import ruamel.yaml


fake = Faker()

pendingToImplement = "Version 0.1 not implemented"
missingEntity = "Missing entity name"

newline = "\n  "
newsubline = "\n   "
propertyTypes = ["Property", "Relationship", "GeoProperty"]

# property check
incompleteDescription = "Incomplete description"
withoutDescription = "No description at all"
wrongTypeDescription = "Wrong NGSI types"
missingTypeDescription = "Missing NGSI types"
exceptions = ["coordinates", "bbox", "type"]

KEYWORDS_FOR_CERTAIN_CHECK = "smart-data-models"
CHECKED_PROPERTY_CASES = ['well documented', 'already used', 'newly available', 'Metadata', 'Failed']

# Regular expression pattern to extract owner, repository, branch, and file path
github_url_pattern = r"https://github\.com/([^/]+)/([^/]+)/blob/([^/]+)/(.+)"

def is_url_existed(url):
    """
        Check if a URL exists and retrieve its content.
        Parameters:
        - url (str): The URL to check.
        Returns:
        - list[bool, str]: A list containing a boolean indicating if the URL exists,
          and either the content of the URL if it exists, or the HTTP status code
          if it doesn't. If there's an exception during the request, it returns
          a boolean indicating failure and an error message.
        Example:
            is_url_existed("https://example.com")
        [True, 'Content of the URL']
            is_url_existed("https://nonexistent-url.com")
        [False, 404]
            is_url_existed("invalid-url")
        [False, 'wrong domain']
        """

    try:
        pointer = requests.get(url)
        if pointer.status_code == 200:
            return [True, pointer.text]
        else:
            return [False, pointer.status_code]
    except:
        return [False, "wrong domain"]


def open_json(fileUrl: str):
    """
    Opens a JSON file given its URL or path and returns the loaded content as a JSON object.
    Parameters:
    - file_url (str): The URL or path of the JSON file.
    Returns:
    - dict: The loaded JSON content if successful, none otherwise.
    Example:
    >>> open_json("https://example.com/data.json")
    {...}
    >>> open_json("local_file.json")
    {...}
    >>> open_json("invalid-url")
    None
    """
    if fileUrl[0:4] == "http":
        # es URL
        try:
            pointer = requests.get(fileUrl)
            return json.loads(pointer.content.decode('utf-8'))
        except:
            return None
    else:
        # es file
        try:
            file = open(fileUrl, "r")
            return json.loads(file.read())
        except:
            return None


def open_jsonref(fileUrl: str):
    """
    Opens a JSON file given its URL or path and returns the loaded content as a JSON object.
    Capable of parsing JSON file with $ref
    Parameters:
    - file_url (str): The URL or path of the JSON file.
    Returns:
    - dict: The loaded JSON content if successful, none otherwise.
    Example:
    >>> open_jsonref("https://example.com/data.json")
    {...}
    >>> open_jsonref("local_file.json")
    {...}
    >>> open_jsonref("invalid-url")
    None
    """
    if fileUrl[0:4] == "http":
        # es URL
        try:
            pointer = requests.get(fileUrl)
            output = jsonref.loads(pointer.content.decode('utf-8'), load_on_repr=True, merge_props=True)
            return output
        except:
            return None
    else:
        # es file
        try:
            file = open(fileUrl, "r")
            return jsonref.loads(file.read(), load_on_repr=True, merge_props=True)
        except:
            return None


def open_yaml(file_url):
    """
    Opens a YAML file either from a URL or a local file path and returns its content as a dictionary.
    """
    yaml = ruamel.yaml.YAML(typ='safe')
    try:
        _, file_extension = os.path.splitext(file_url)
        if file_extension.lower() != '.yaml':
            print("Invalid file format. The file should have a .yaml extension.")

        if file_url.startswith("http"):
            if re.match(github_url_pattern, file_url):
                raw_github_url = convert_to_raw_github_url(file_url)
                response = requests.get(raw_github_url)
                response.raise_for_status()  # Raise HTTPError for bad responses
                return yaml.load(response.content.decode('utf-8'))
            else:
                response = requests.get(file_url)
                response.raise_for_status()
                return yaml.load(response.content.decode('utf-8'))
        else:
            with open(file_url, "r") as file:
                return yaml.load(file)
    except requests.exceptions.RequestException as e:
        print(f"Failed to fetch content from URL: {e}")
    except Exception as e:
        print(f"Error: {e}")
        

def extract_datamodel_from_raw_url(schemaUrl: str) -> str:
    """
    Extracts the name of the data model from the given schema repository URL.
    Parameters:
    - schema_url (str): The URL of the schema repository.
    Returns:
    - str: The name of the data model extracted from the URL. Return 'dataModel' as a flag naming doesn't find any match
    Example:
    >>> extract_datamodel_from_raw_url("https://raw.githubusercontent.com/smart-data-models/dataModel.WasteWater/master/WaterProcess/schema.json")
    'WaterProcess'
    >>> extract_datamodel_from_raw_url("https://smart-data-models.github.io/dataModel.WasteWater/WaterProcess/schema.json")
    'WaterProcess'
    >>> extract_datamodel_from_raw_url("invalid-url")
    'dataModel'
    """
    import re

    patternRaw = "https:\/\/raw\.githubusercontent\.com\/smart\-data\-models\/.*\/(.*)\/schema\.json"
    patternPage = "https:\/\/smart-data-models\.github\.io\/.*\/(.*)\/schema\.json"
    urlSearch = re.search(patternRaw, schemaUrl, re.IGNORECASE)

    url2Search = re.search(patternPage, schemaUrl, re.IGNORECASE)
    if urlSearch:
        dataModel = urlSearch.group(1)
    elif url2Search:
        dataModel = url2Search.group(1)
    else:
        dataModel = "dataModel"
    return dataModel

def extract_subject_from_raw_url(schemaUrl: str) -> str:
    """
    Extracts the name of the subject from the given schema repository URL.
    Parameters:
    - schema_url (str): The URL of the schema repository.
    Returns:
    - str: The name of the subject extracted from the URL. Return 'subject' as a flag naming doesn't find any match
    Example:
    >>> extract_datamodel_from_raw_url("https://raw.githubusercontent.com/smart-data-models/dataModel.WasteWater/master/WaterProcess/schema.json")
    'dataModel.WasteWater'
    >>> extract_datamodel_from_raw_url("https://smart-data-models.github.io/dataModel.WasteWater/WaterProcess/schema.json")
    'dataModel.WasteWater'
    >>> extract_datamodel_from_raw_url("invalid-url")
    'subject'
    """
    import re
    patternRaw = "https:\/\/raw\.githubusercontent\.com\/smart\-data\-models\/(.*)\/.*\/schema\.json"
    patternPage = "https:\/\/smart-data-models\.github\.io\/(.*)\/.*\/schema\.json"
    urlSearch = re.search(patternRaw, schemaUrl, re.IGNORECASE)
    url2Search = re.search(patternPage, schemaUrl, re.IGNORECASE)
    if urlSearch:
        subject = urlSearch.group(1).split("/")[0]
    elif url2Search:
        subject = url2Search.group(1).split("/")[0]
    else:
        subject = "subject"
    return subject


def create_context(subject: str) -> str:
    """
    Creates the link to context.jsonld of given subject

    Parameters:
    - subject (str): The name of the subject.

    Returns:
    - str: The link of the context.jsonld

    Example:
    >>> create_context("dataModel.WasteWater")
    'https://smart-data-models.github.io/dataModel.WasteWater/context.jsonld'
    """
    return "https://smart-data-models.github.io/" + subject + "/context.jsonld"

#########################################
#   Conversion between different examples
#########################################

@DeprecationWarning
def normalized2keyvalues(normalizedPayload: dict) -> dict:
    """
    Converts the normalized format example to key-value format example
    Parameters:
    - normalizedPayload (str): The JSON payload of the normalized format example.
    Returns:
    - dict: The JSON payload of the key-value format example
    Example:
       normalized2keyvalues(
        {
            "id": "urn:waterdna:haemil:WaterProcess_01",
            "type": "WaterProcess",
            "location": {
                "type": "geo:json",
                "value": {
                    "type": "Point",
                    "coordinates": [
                            127.266663,
                            36.524677
                        ]
                    }
            },
            "name": {
                "type": "Text",
                "value": "inflow process at water treatment plant 1"
            },
            ...
        }
    )
    {
        "id": "urn:waterdna:haemil:WaterProcess_01",
        "type": "WaterProcess",
        "location": {
            "type": "Point",
            "coordinates": [
                    127.266663,
                    36.524677
                ]
        },
        "name": "inflow process at water treatment plant 1",
        ...
    }
    """

    normalizedDict = normalizedPayload
    output = {}

    for element in normalizedDict:
        try:
            # In a normalized format, `type` and `value` are two mandatory keys
            # To get the key-value format, simply by taking the value inside `value`
            if "value" in normalizedDict[element]:
                value = normalizedDict[element]["value"]
                # In NGSI-LD normalized format, `@type` and `@value` are used to represent
                # DateTime as a sub-structure inside `value`
                # If `@type` exists, then take the value inside `@value`
                if "@type" in value:
                    output[element] = normalizedDict[element]["value"]["@value"]
                else:
                    output[element] = value
            else:
                output[element] = normalizedDict[element]
        except:
            output[element] = normalizedDict[element]

    return output

def normalized2keyvalues_v2(normalizedPayload: dict, level: int = 0) -> dict:
    """
    Converts the normalized format example to key-value format example, version 2
    Could be applied to normalized.json or normalized.jsonld

    Parameters:
    - normalizedPayload (str): The JSON payload of the normalized format example.

    Returns:
    - dict: The JSON payload of the key-value format example
    """
    output = {}
    for element, body in normalizedPayload.items():

        # If the body is a list of dictionary
        # recursively process the list, get the sub-structure of `value` if it exists
        if isinstance(body, list) and body and isinstance(body[0], dict):
            output[element] = [normalized2keyvalues_v2(subprop, level + 1).get("value", subprop) for subprop in body]

        # If the body is a dictionary, then process it in detail
        elif isinstance(body, dict):

            # Get the sub-structure of `@value` if it exists
            # This refers to the case of DateTime in NGSI-LD
            # Example: {"type": "Property", "value": {"@type": "DateTime", "@value": "xxx"}}
            if "@value" in body:
                output[element] = body["@value"]

            elif "value" in body:
                value = body["value"]

                # If the value is a dictionary, then recursively process the value
                if isinstance(value, dict):
                    output[element] = normalized2keyvalues_v2({"value": value}, level + 1)["value"]

                # If the value is a list of dictionary
                # then recursively process the list, get the sub-structure of `value` if it exists
                elif isinstance(value, list) and value and isinstance(value[0], dict):
                    output[element] = [normalized2keyvalues_v2(subvalue, level + 1).get("value", subvalue) for subvalue
                                       in value]

                # Otherwise, use the value as is
                else:
                    output[element] = value

            # Get the sub-structure of `object` if it exists
            # This refers to the case of Relationship in NGSI-LD
            # Example: {"type": "Relationship", "object": "xxxx"}
            elif "object" in body:
                output[element] = body["object"]

            # Otherwise, use the body as is
            else:
                output[element] = body

        # Otherwise, use the body as is
        else:
            output[element] = body

    return output

def keyvalues2normalized_ngsild(keyvaluesPayload: dict, yamlDict: dict, detailed: bool = True, level: int = 0) -> dict:
    """
    Converts the key-value format example to normalized format example in NGSI-LD

    Parameters:
    - keyvaluesPayload (dict): The JSON payload of the key-value format example.
    - yamlDict (dict): The schema dictionary.
    - detailed (bool): Whether generate only the first-level properties or for all in detailed
    - level (int): Level of recursion.

    Returns:
    - dict: The JSON payload of the normalized.jsonld format example
    """
    import json

    def valid_date(datestring: str) -> tuple[bool, str]:
        """
        Valid the date string and return the specific type according to the patterns, "Date", "DateTime", "Time", "Text"
        """
        import re
        from datetime import time

        if not ("T" in datestring) and ("Z" in datestring):
            try:
                time.fromisoformat(datestring.replace('Z', '+00:00'))
            except:
                return False, "Text"
            return True, "Time"

        else:
            date = datestring.split("T")[0]
            try:
                validDate = re.match('^[0-9]{2,4}[-/][0-9]{2}[-/][0-9]{2,4}$', date)
            except ValueError:
                return False, "Text"

            if validDate is not None:
                if len(datestring.split("T")) > 1:
                    return True, "DateTime"
                return True, "Date"
            else:
                return False, "Text"

    keyvaluesDict = keyvaluesPayload
    output = {}
    for element in keyvaluesDict:

        # id, type, @context, are in key-value format by default
        if level == 0 and element in ["id", "type", "@context"]: continue
        item = {}

        if isinstance(keyvaluesDict[element], list):
            # it is an array
            item["type"] = yamlDict[element]['x-ngsi']['type']
            if detailed:
                if len(keyvaluesDict[element]) > 0 and isinstance(keyvaluesDict[element][0], dict):
                    tmpList = []
                    for idx in range(len(keyvaluesDict[element])):
                        tmpList.append(keyvalues2normalized_ngsild(keyvaluesDict[element][idx], yamlDict[element][idx],
                                                                   level=level + 1))
                    item["value"] = tmpList
                else:
                    item["value"] = keyvaluesDict[element]
            else:
                item["value"] = keyvaluesDict[element]

        elif isinstance(keyvaluesDict[element], dict):
            # it is an object
            # item["type"] = "object"
            if element == "location":
                item["type"] = yamlDict[element]['x-ngsi']['type']
            elif "type" in keyvaluesDict[element] and "coordinates" in keyvaluesDict[element]:  # location-like property
                item["type"] = yamlDict[element]['x-ngsi']['type']
            else:
                item["type"] = yamlDict[element]['x-ngsi']['type']
            if detailed:
                item["value"] = keyvalues2normalized_ngsild(keyvaluesDict[element], yamlDict[element], level=level + 1)
            else:
                item["value"] = keyvaluesDict[element]

        elif isinstance(keyvaluesDict[element], str):
            dateFlag, dateType = valid_date(keyvaluesDict[element])
            if dateFlag:
                # it is a date
                item["type"] = yamlDict[element]['x-ngsi']['type']
                if dateType == "Date":
                    item["value"] = {"@type": "Date", "@value": keyvaluesDict[element]}
                elif dateType == "Time":
                    item["value"] = {"@type": "Time", "@value": keyvaluesDict[element]}
                else:
                    item["value"] = {"@type": "DateTime", "@value": keyvaluesDict[element]}
            else:
                # it is a string
                item["type"] = yamlDict[element]['x-ngsi']['type']
                item["value"] = keyvaluesDict[element]

        elif isinstance(keyvaluesDict[element], int) or isinstance(keyvaluesDict[element], float):
            # it is a number
            item["type"] = yamlDict[element]['x-ngsi']['type']
            item["value"] = keyvaluesDict[element]

        elif keyvaluesDict[element] == True:
            # it is a boolean for true
            item["type"] = yamlDict[element]['x-ngsi']['type']
            item["value"] = json.loads("true")

        elif keyvaluesDict[element] == False:
            # it is a boolean for false
            item["type"] = yamlDict[element]['x-ngsi']['type']
            item["value"] = json.loads("false")

        else:
            print("*** other type ***")
            print("I do not know what is it")
            print(keyvaluesDict[element])
            print("--- other type ---")
        output[element] = item

    if "id" in keyvaluesDict:
        output["id"] = keyvaluesDict["id"]
    if "type" in keyvaluesDict:
        output["type"] = keyvaluesDict["type"]
    if "@context" in keyvaluesDict:
        output["@context"] = keyvaluesDict["@context"]

    return output

def keyvalues2normalized_ngsiv2(keyvaluesPayload, detailed=True):
    """
    Converts the key-value format example to normalized format example in NGSI-V2

    Parameters:
    - keyvaluesPayload (dict): The JSON payload of the key-value format example.
    - detailed (bool): Whether generate only the first-level properties or for all in detailed

    Returns:
    - dict: The JSON payload of the normalized.json format example
    """
    import json

    def valid_date(datestring: str) -> bool:
        """
        Valid the date string and return the specific type according to the patterns, "DateTime", "Text", only two types in NGSI-V2
        """
        import re
        date = datestring.split("T")[0]
        try:
            validDate = re.match('^[0-9]{2,4}[-/][0-9]{2}[-/][0-9]{2,4}$', date)
        except ValueError:
            return False

        if validDate is not None:
            return True
        else:
            return False

    keyvaluesDict = keyvaluesPayload
    output = {}

    for element in keyvaluesDict:
        item = {}

        if isinstance(keyvaluesDict[element], list):
            # it is an array
            # item["type"] = "array"
            item["type"] = "StructuredValue"
            if detailed:
                if len(keyvaluesDict[element]) > 0 and isinstance(keyvaluesDict[element][0], dict):
                    tmpList = []
                    for idx in range(len(keyvaluesDict[element])):
                        tmpList.append(keyvalues2normalized_ngsiv2(keyvaluesDict[element][idx]))
                    item["value"] = tmpList
                else:
                    item["value"] = keyvaluesDict[element]
            else:
                item["value"] = keyvaluesDict[element]

        elif isinstance(keyvaluesDict[element], dict):
            # it is an object
            if element == "location":
                item["type"] = "geo:json"
            else:
                item["type"] = "StructuredValue"
            if detailed:
                item["value"] = keyvalues2normalized_ngsiv2(keyvaluesDict[element])
            else:
                item["value"] = keyvaluesDict[element]

        elif isinstance(keyvaluesDict[element], str):
            if valid_date(keyvaluesDict[element]):
                # it is a date
                item["type"] = "DateTime"
            else:
                # it is a string
                item["type"] = "Text"
            item["value"] = keyvaluesDict[element]

        elif isinstance(keyvaluesDict[element], int) or isinstance(keyvaluesDict[element], float):
            # it is a number
            item["type"] = "Number"
            item["value"] = keyvaluesDict[element]

        elif keyvaluesDict[element] == True:
            # it is a boolean
            item["type"] = "Boolean"
            item["value"] = json.loads("true")

        elif keyvaluesDict[element] == False:
            # it is a boolean
            item["type"] = "Boolean"
            item["value"] = json.loads("false")

        else:
            print("*** other type ***")
            print("I do now know what is it")
            print(keyvaluesDict[element])
            print("--- other type ---")
        output[element] = item

    if "id" in output:
        output["id"] = output["id"]["value"]
    if "type" in output:
        output["type"] = output["type"]["value"]
    if "@context" in output:
        output["@context"] = output["@context"]["value"]

    return output

def generate_random_string(length=4) -> str:
    """
    Generates a random string of the specified length.
    Parameters:
    - length (int): The length of the random string. Default is 4.
    Returns:
    - str: The generated random string.
    Example:
    >>> generate_random_string()
    'aB3x'
    >>> generate_random_string(8)
    'y2RtL6zA'
    """
    characters = string.ascii_letters + string.digits
    return ''.join(random.choice(characters) for _ in range(length))

########################################
#   Generate fake example related
########################################

def fake_uri(propertyName:str, dataModel:str) -> str:
    """
        Generates a fake URI based on the provided property name and data model.
        Parameters:
        - propertyName (str): The name of the property.
        - dataModel (str): The name of the data model.
        Returns:
        - str: The generated fake URI.
        Example:
        >>> fake_uri("temperature", "WeatherObserved")
        'urn:ngsi-ld:WeatherObserved:temperature:ABCD1234'
        """
    output = "urn:ngsi-ld:" + str(dataModel) + ":" + str(propertyName) + ":"
    output += fake.bothify(text="????:########", letters="ABCDEFGHIJKLMNOPQRSTUVWXYZ")
    # genera example uri
    return output

def payload_uri(propertyName: str, dataModel: str) -> dict:
    """
       Generates a payload containing a URI value for a Property.
       Parameters:
       - propertyName (str): The name of the property.
       - dataModel (str): The name of the data model.
       Returns:
       - dict: A payload containing a URI value.
       Example:
       >>> payload_uri("temperature", "WeatherObserved")
       {'type': 'Property', 'value': 'urn:ngsi-ld:WeatherObserved:temperature:ABCD1234'}
       """
    return {"type": "Property", "value": fake_uri(propertyName, dataModel)}

def payload_relationship(propertyName, dataModel):
    """
        Generates a payload containing a URI value for a Relationship.
        Parameters:
        - propertyName (str): The name of the relationship property.
        - dataModel (str): The name of the data model.
        Returns:
        - dict: A payload containing a relationship property.
        Example:
        >>> payload_relationship("belongsTo", "Building")
        {'type': 'Relationship', 'object': 'urn:ngsi-ld:Building:belongsTo:ABCD1234'}
        TODO
        Copy with different data structure relationships, like list and dict
        """
    return {"type": "object", "value": fake_uri(propertyName, dataModel)}


def fake_geoproperty(geo_type: str) -> dict:
    """
    Generates a fake GeoJSON property based on the given geometry type.

    Parameters:
    - geo_type (str): The type of the geometry (e.g., "Point", "LineString", "Polygon").

    Returns:
    - dict: A GeoJSON property with fake coordinates.

    Example:
    >>> fake_geoproperty("Point")
    {'type': 'Point', 'coordinates': [12.345, -67.890]}

    TODO
    To support more geo_types
    """
    dict_value = {}

    if geo_type == "Point":
        dict_value = {"type": geo_type, "coordinates": [float(fake.latitude()), float(fake.longitude())]}
    elif geo_type == "LineString":
        dict_value = {"type": geo_type, "coordinates": [
            [float(fake.latitude()), float(fake.longitude())],
            [float(fake.latitude()), float(fake.longitude())]
        ]}
    elif geo_type == "Polygon":
        dict_value = {"type": geo_type, "coordinates": [
            [float(fake.latitude()), float(fake.longitude())],
            [float(fake.latitude()), float(fake.longitude())],
            [float(fake.latitude()), float(fake.longitude())],
            [float(fake.latitude()), float(fake.longitude())]
        ]}

    return dict_value

def payload_geoproperty(geo_type: str) -> dict:
    """
    Generates a GeoJSON property payload with a fake geometry based on the given type.

    Parameters:
    - geo_type (str): The type of the geometry (e.g., "Point", "LineString", "Polygon").

    Returns:
    - dict: A GeoJSON property payload.

    Example:
    >>> payload_geoproperty("Point")
    {'type': 'GeoProperty', 'value': {'type': 'Point', 'coordinates': [12.345, -67.890]}}
    """
    return {"type": "GeoProperty", "value": fake_geoproperty(geo_type)}


def fake_number(*args) -> float:
    """
    Generates a fake floating-point number based on the provided arguments.

    Parameters:
    - If only one argument is provided (an integer), it represents the number of left digits (default is 6).
    - If more than one argument is provided as a tuple, the first element is the number of left digits (default is 6),
      the second element is the minimum value (default is 0), and the third element is the maximum value
      (default is maximum of 1000 or minimum + 1000).

    Returns:
    - float: A fake floating-point number.

    Example:
    >>> fake_number()
    123.4

    >>> fake_number(8)
    12345678.9

    >>> fake_number((8, 100, 500))
    123456.7
    """
    Faker.seed()

    if isinstance(args[0], int):
        left_digits = 6
        minimum = 0
        maximum = 1000
    else:
        left_digits = int(args[0][0])
        minimum = int(args[0][1]) if len(args[0]) > 1 else 0
        maximum = int(args[0][2]) if len(args[0]) > 2 else max(1000, minimum + 1000)

    return fake.pyfloat(left_digits=left_digits, right_digits=1, min_value=minimum, max_value=maximum)

def fake_integer(*args) -> int:
    """
    Generates a fake integer based on the provided arguments.

    Parameters:
    - If only one argument is provided (an integer), it represents the maximum value (default is 1000).
    - If more than one argument is provided as a tuple, the first element is the minimum value (default is 0),
      and the second element is the maximum value (default is maximum of 1000 or minimum + 1000).

    Returns:
    - int: A fake integer.

    Example:
    >>> fake_integer()
    42

    >>> fake_integer(50)
    25

    >>> fake_integer((10, 50))
    37
    """
    Faker.seed()

    if isinstance(args[0], int):
        minimum = 0
        maximum = args[0]
    else:
        minimum = int(args[0][0]) if len(args[0]) > 0 else 0
        maximum = int(args[0][1]) if len(args[0]) > 1 else max(1000, minimum + 1000)

    return fake.pyint(minimum, maximum)

def payload_number(*args) -> dict:
    """
    Generates a payload for a fake number property based on the provided arguments.

    Parameters:
    - If only one argument is provided (an integer), it represents the maximum value (default is 1000).
    - If more than one argument is provided as a tuple, the first element is the number of left digits (default is 6),
      the second element is the minimum value (default is 0), and the third element is the maximum value
      (default is maximum of 1000 or minimum + 1000).

    Returns:
    - dict: A payload dictionary with the type 'Property' and a fake number value.

    Example:
    >>> payload_number()
    {'type': 'Property', 'value': 42.3}

    >>> payload_number(50)
    {'type': 'Property', 'value': 25.7}

    >>> payload_number((6, 10, 50))
    {'type': 'Property', 'value': 37.9}
    """
    return {"type": "Property", "value": fake_number(*args)}

def payload_integer(*args) -> dict:
    """
    Generates a payload for a fake integer property based on the provided arguments.

    Parameters:
    - If only one argument is provided (an integer), it represents the maximum value (default is 1000).
    - If more than one argument is provided as a tuple, the first element is the minimum value (default is 0),
      and the second element is the maximum value (default is maximum of 1000 or minimum + 1000).

    Returns:
    - dict: A payload dictionary with the type 'Property' and a fake integer value.

    Example:
    >>> payload_integer()
    {'type': 'Property', 'value': 42}

    >>> payload_integer((10, 50))
    {'type': 'Property', 'value': 37}
    """
    return {"type": "Property", "value": fake_integer(*args)}

def fake_datetime() -> str:
    """
    Generates a fake datetime string in the format '%Y-%m-%dT%H:%M:%SZ'.

    Returns:
    - str: A fake datetime string.

    Example:
    >>> fake_datetime()
    '2023-05-16T14:23:45Z'
    """
    return fake.date(pattern='%Y-%m-%dT%H:%M:%SZ', end_datetime=None)


def fake_date() -> str:
    """
    Generates a fake date string in the format '%Y-%m-%d'.

    Returns:
    - str: A fake date string.

    Example:
    >>> fake_date()
    '2023-05-16'
    """
    return fake.date(pattern='%Y-%m-%d', end_datetime=None)

def fake_time() -> str:
    """
    Generates a fake time string in the format '%H:%M:%SZ'.

    Returns:
    - str: A fake time string.

    Example:
    >>> fake_time()
    '14:23:45Z'
    """
    return fake.date(pattern='%H:%M:%SZ', end_datetime=None)

def payload_datetime() -> dict:
    """
    Generates a payload for a fake datetime property.

    Returns:
    - dict: A payload dictionary with the type 'Property' and a fake datetime value.

    Example:
    >>> payload_datetime()
    {'type': 'Property', 'value': {'@type': 'DateTime', '@value': '2023-05-16T14:23:45Z'}}
    """
    return {"type": "Property", "value": {"@type": "DateTime", "@value": fake_datetime()}}

def payload_date() -> dict:
    """
    Generates a payload for a fake date property.

    Returns:
    - dict: A payload dictionary with the type 'Property' and a fake date value.

    Example:
    >>> payload_date()
    {'type': 'Property', 'value': {'@type': 'Date', '@value': '2023-05-16'}}
    """
    return {"type": "Property", "value": {"@type": "Date", "@value": fake_date()}}

def payload_time() -> dict:
    """
    Generates a payload for a fake time property.

    Returns:
    - dict: A payload dictionary with the type 'Property' and a fake time value.

    Example:
    >>> payload_time()
    {'type': 'Property', 'value': {'@type': 'Time', '@value': '14:23:45Z'}}
    """
    return {"type": "Property", "value": {"@type": "Time", "@value": fake_time()}}

def fake_email(dataModel: str) -> str:
    """
    Generates a fake email address based on the data model.

    Parameters:
    - dataModel (str): The data model to incorporate into the email address.

    Returns:
    - str: A fake email address.

    Example:
    >>> fake_email('SmartCity')
    'john.doe@smartcity.com'
    """
    list_of_domains = (
        'com',
        'net',
        'org',
        'gov'
    )

    dns_org = fake.random_choices(
        elements=list_of_domains,
        length=1
    )[0]

    return f"{fake.first_name()}.{fake.last_name()}@{dataModel}.{dns_org}".lower()

def payload_email(dataModel: str) -> dict:
    """
    Generates a payload for a fake email property.

    Parameters:
    - dataModel (str): The data model to incorporate into the email address.

    Returns:
    - dict: A payload dictionary with the type 'Property' and a fake email value.

    Example:
    >>> payload_email('SmartCity')
    {'type': 'Property', 'value': 'john.doe@smartcity.com'}
    """
    return {"type": "Property", "value": fake_email(dataModel)}


def fake_string(length: int, min_length: int, max_length:int) -> str:
    """
    Generate a fake string based on specified length and range.

    Parameters:
    - length (int): Desired length of the string.
    - min_length (int): Minimum possible length of the string.
    - max_length (int): Maximum possible length of the string.

    Returns:
    - str: Fake string.

    Example:
    >>> fake_string(1, 5, 10)
    'Lorem ipsum'
    """
    paragraph_length = random.randint(min_length, max_length)

    # generate example string (regular)
    if length == 1:
        return fake.paragraphs(nb=length)[0][:paragraph_length]
    else:
        return fake.paragraphs(nb=length)

def payload_string(length: int, min_length: int, max_length: int, pattern: str="") -> dict:
    """
    Generate a payload for a string property with optional pattern.

    Parameters:
    - length (int): Desired length of the string.
    - min_length (int): Minimum possible length of the string.
    - max_length (int): Maximum possible length of the string.
    - pattern (str, optional): Regular expression pattern to match.

    Returns:
    - dict: Property payload.

    Example:
    >>> payload_string(1, 5, 10)
    {'type': 'Property', 'value': 'Lorem ipsum'}
    """
    if pattern:
        return {"type": "Property", "value": rstr.xeger(pattern)}
    else:
        return {"type": "Property", "value": fake_string(length, min_length, max_length)}


def fake_enum(elements_list: list, length: int) -> str:
    """
    Generates a fake enum.

    Parameters:
    - elements_list (list): The list of possible enum values.
    - length (int): The number of enum values to generate.

    Returns:
    - str: A fake enum value or a list of fake enum values.

    Example:
    >>> fake_enum(['red', 'blue', 'green'], 1)
    'red'
    """
    if length == 1:
        return fake.random_choices(elements=elements_list, length=length)[0]
    else:
        return fake.random_choices(elements=elements_list, length=length)

def payload_enum(elements_list: list, length: int) -> dict:
    """
    Generates a payload for a fake enum property.

    Parameters:
    - elements_list (list): The list of possible enum values.
    - length (int): The number of enum values to generate.

    Returns:
    - dict: A payload dictionary with the type 'Property' and a fake enum value or values.

    Example:
    >>> payload_enum(['red', 'blue', 'green'], 3)
    {'type': 'Property', 'value': ['red', 'blue', 'green']}
    """
    return {"type": "Property", "value": fake_enum(elements_list, length)}


def fake_boolean() -> bool:
    """
    Generates a fake boolean value.

    Returns:
    - bool: A random boolean value.

    Example:
    >>> fake_boolean()
    True
    """
    return fake.random_choices(elements=[True, False], length=1)[0]

def payload_boolean() -> dict:
    """
    Generates a payload for a fake boolean property.

    Returns:
    - dict: A payload dictionary with the type 'Property' and a fake boolean value.

    Example:
    >>> payload_boolean()
    {'type': 'Property', 'value': True}
    """
    return {"type": "Property", "value": fake_boolean()}


def parse_schema_description(description) -> Tuple[Dict, str]:
    """
    Parses the schema description and extracts relevant information.

    Parameters:
    - description (str): The description string to be parsed.

    Returns:
    - tuple[dict, str]: A tuple contains a dictionary of NGSI type metadata and description.

    Example:
    >>> parse_schema_description("Property. This is a xxx. Model: 'TemperatureSensorModel'. Enum: 'High', 'Low'. Privacy: 'High'")
    [{"type": "Property", "model": "TemperatureSensorModel", "enum": ['High', 'Low'], "privacy": "High"}, "This is a xxx."]
    """
    ngsiOutput = {}

    # Process the description: remove the double-quotes, add spaces between sentences
    description = description.replace(chr(34), "")
    description = re.sub(r'\.([A-Z])', r'. \1', description)

    description = description.split(". ")
    copiedDescription = list.copy(description)

    for item in description:
        if item in propertyTypes:
            ngsiOutput["type"] = item
            copiedDescription.remove(item)
        elif item.find("Model:") > -1:
            copiedDescription.remove(item)
            ngsiOutput["model"] = (item.replace("'", ""))[len("Model:"):]
        elif item.find("Units:") > -1:
            copiedDescription.remove(item)
            ngsiOutput["units"] = (item.replace("'", ""))[len("Units:"):]
        elif item.find("Enum:") > -1:
            copiedDescription.remove(item)
            raw = (item.replace("'", ""))[len("Enum:"):]
            ngsiOutput["enum"] = raw.split(", ")
        elif item.find("Privacy:") > -1:
            copiedDescription.remove(item)
            ngsiOutput["privacy"] = (item.replace("'", ""))[len("Privacy:"):]

    description = ". ".join(copiedDescription)

    return ngsiOutput, description


def echo(concept: str, variable) -> None:
    """
    Print a formatted representation of a variable.

    Parameters:
    - concept (str): A string describing the concept of the variable.
    - variable: The variable to be printed.

    Example:
    >>> echo("Title", "Hello, World!")
    *** Title ***
    Hello, World!
    --- Title ---
    """
    print("*** " + concept + " ***")
    print(variable)
    print("--- " + concept + " ---")


def fake_array(type: str, numItems: int, options=None) -> list:
    """
    Generate an array of fake data based on the specified type and options.

    Parameters:
    - type (str): Type of elements in the array (e.g., "number", "boolean", "string", "date-time", "enum", "array", "object").
    - numItems (int): Number of items to generate in the array.
    - options (tuple, int, list, or dict): Options specific to the chosen type.

    Returns:
    - list: Array of fake data.

    Example:
    >>> fake_array("number", 3, (6, 0, 100))
    [45.2, 18.6, 97.0]

    TODO
    To be implemented based on the type and options of the items in the array
    To be implemented based on the structure of the object
    """
    if type == "number":
        return [fake_number(options) for _ in range(numItems)]
    elif type == "boolean":
        return [fake_boolean() for _ in range(numItems)]
    elif type == "string":
        return [fake_string(options) for _ in range(numItems)]
    elif type == "date-time":
        return [fake_datetime() for _ in range(numItems)]
    elif type == "enum":
        return [fake_enum(options) for _ in range(numItems)]
    elif type == "array":
        return []
    elif type == "object":
        return []

def payload_array(type: str, numItems: int, options) -> dict:
    """
    Generate a payload for an array property.

    Parameters:
    - type (str): Type of elements in the array ("number", "boolean", "string", "date-time", "enum", "array", "object").
    - numItems (int): Number of items to generate in the array.
    - options (tuple, int, list, or dict): Options specific to the chosen type.
        options for type number is a tuple with (length, minimum and maximum)
        options for type string is an integer with the length in paragraphs
        options for type enum is the list with the possible values
        options for array is a list with two values [type for the items, options for this type of item]
        options for object is the complete payload

    Returns:
    - dict: Property payload.

    Example:
    >>> payload_array("number", 3, (6, 0, 100))
    {'type': 'Property', 'value': [45.2, 18.6, 97.0]}
    """
    if type is not "array" and type is not "object":
        return {"type": "Property", "value": fake_array(type, numItems, options)}
    else:
        return pendingToImplement


def parse_property(fullPayload, dataModel, level):
    # echo("payload", fullPayload)
    prop = [p for p in fullPayload.keys()][0]
    # echo("prop", prop)
    payload = fullPayload[prop]

    keys = [k for k in payload]
    # echo("keys", keys)
    if "required" in keys:
        del fullPayload[prop]["required"]

    if "description" in keys:
        # [description, Property/Relationship/GeoProperty, Model, Units, Enum, Privacy]
        metadata = parse_schema_description(payload["description"])
    else:
        metadata = ["", "Property"]
    if prop == "id" and level == 0:
        return {"id": fake_uri("id", dataModel)}
    elif prop == "id" and level > 0:
        return payload_uri("id", "id")
    elif "patternProperties" in keys:
        # print("Detected patternProperties")
        keyPatterned = list(payload["patternProperties"].keys())[0]
        propertyName = rstr.xeger(keyPatterned)
        # print(keyPatterned, propertyName)
        # print(fullPayload)
        if "required" in fullPayload[prop]:
            del fullPayload[prop]["required"]
        return parse_property({propertyName: fullPayload[prop]["patternProperties"][keyPatterned]}, dataModel, level+1)
    elif prop == "type" and level == 0:
        if "enum" in fullPayload[prop]:
            return fullPayload[prop]["enum"][0]
        else:
            return missingEntity
    elif prop == "location":
        return payload_geoproperty("Point")
    elif metadata[1] == "Relationship":
        return payload_relationship(prop, dataModel)
    elif metadata[1] == "GeoProperty":
        return payload_geoproperty("Point")
    elif prop in ["anyOf", "allOf", "oneOf"]:
        return parse_property({"item": fullPayload[prop][0]}, dataModel, level)
    elif "type" in keys:
        if payload["type"] == "number":
            argsNumber = (4,)
            ### treating number with/without maximum/minimum ###
            if "minimum" in payload:
                argsNumber = argsNumber + (float(payload["minimum"]),)
            else:
                argsNumber = argsNumber + (0,)
            if "maximum" in payload:
                argsNumber = argsNumber + (float(payload["maximum"]),)
            else:
                argsNumber = argsNumber + (max(1000, argsNumber[1] + 1000),)
            return payload_number(argsNumber)
        elif payload["type"] == "integer":
            argsNumber = ()
            ### treating number with/without maximum/minimum ###
            if "minimum" in payload:
                argsNumber = argsNumber + (int(payload["minimum"]),)
            else:
                argsNumber = argsNumber + (0,)
            if "maximum" in payload:
                argsNumber = argsNumber + (int(payload["maximum"]),)
            else:
                argsNumber = argsNumber + (int(argsNumber[0] + 1000),)
            return payload_integer(argsNumber)
        elif payload["type"] == "string" and "format" in keys:
            if (payload["format"] == "date-time") or (payload["format"] == "date"):
                return payload_date()
            elif payload["format"] == "time":
                return payload_time()
            elif payload["format"] == "uri":
                return payload_uri(prop, dataModel)
            elif (payload["format"] == "idn-email") or (payload["format"] == "email"):
                return payload_email(dataModel)
            elif payload["format"] == "text":
                return payload_string(1)
        elif payload["type"] == "string" and "enum" in keys:
            return payload_enum(payload["enum"], 1)
        elif payload["type"] == "string":
            return payload_string(1)
        elif payload["type"] == "boolean":
            return payload_boolean()
        elif payload["type"] == "array":
            if "minItems" in payload:
                arrayItems = payload["minItems"]
            else:
                arrayItems = 2
            if "maxItems" in payload:
                arrayItems = int((arrayItems + int(payload["maxItems"]))/2)
            else:
                arrayItems = arrayItems
            arrayPayload = payload["items"].copy()
            if "required" in arrayPayload:
                del arrayPayload["required"]
            valuesArray = [parse_property({"items": arrayPayload}, dataModel, level+1)["value"] for i in range(arrayItems)]
            return {"type": "Property", "value": valuesArray}
        elif payload["type"] == "object":
            # print("parsing object")
            valuesObject = {}
            # print(payload)
            objectPayload = payload["properties"]
            if "required" in objectPayload:
                del objectPayload["required"]
            for p in objectPayload:
                # echo("{prop: payload[properties]}", {p: payload["properties"][p]})
                if p != "required":
                    valuesObject[p] = parse_property({p: payload["properties"][p]}, dataModel, level+1)["value"]

            return {"type": "Property", "value": valuesObject}
    elif "oneOf" in keys:
        return parse_property({prop: fullPayload[prop]["oneOf"][0]}, dataModel, level)
    else:
        return {prop: pendingToImplement, "value": fake_uri(prop, dataModel)}

def parse_property2ngsild_example(fullPayload: dict, dataModel: str, level: int = 0) -> dict:
    """
    Recursively parse a JSON schema property and generate fake data in NGSI-LD format.

    Parameters:
    - fullPayload (dict): Full payload of the property.
    - dataModel (str): Name of the data model.
    - level (int): Level of recursion.

    Returns:
    - dict: Parsed property with generated fake data.

    Example:
    >>> parse_property2ngsild_example({"id": {"type": "string"}}, "ExampleDataModel", 0)
    {'id': {'type': 'Property', 'value': 'urn:ngsi-ld:ExampleDataModel:id:?????########'}}

    """
    prop = [p for p in fullPayload.keys()][0]
    payload = fullPayload[prop]

    # Get all the keys within this property
    keys = [k for k in payload]

    # Ignore `required` so far
    if "required" in keys:
        del fullPayload[prop]["required"]

    if "description" in keys:
        # parse description of the property
        # return a list, [description, Property/Relationship/GeoProperty, Model, Units, Enum, Privacy]
        ngsiOutput, description = parse_schema_description(payload["description"])
        metadata = [description, ngsiOutput["type"]]
    else:
        metadata = ["", "Property"]

    if prop == "id" and level == 0:
        # first level `id` is a system-defined property, displayed in key-value format by default
        return {"id": fake_uri("id", dataModel)}

    elif prop == "id" and level > 0:
        return payload_uri("id", dataModel)

    elif "patternProperties" in keys:
        # if `patternProperties` is defined, use the first pattern to generate
        keyPatterned = list(payload["patternProperties"].keys())[0]
        propertyName = rstr.xeger(keyPatterned)
        if "required" in fullPayload[prop]:
            del fullPayload[prop]["required"]
        return parse_property2ngsild_example({propertyName: fullPayload[prop]["patternProperties"][keyPatterned]},
                                             dataModel, level + 1)

    elif prop == "type" and level == 0:
        # first level `type` is a system-defined property, displayed in key-value format by default
        # `enum` is a key inside `type`, store the name of the data model
        # it is an error if `enum` is not found
        if "enum" in fullPayload[prop]:
            return fullPayload[prop]["enum"][0]
        else:
            return missingEntity

    elif prop == "location" or metadata[1] == "GeoProperty":
        # `location` or `GeoProperty`-like is displayed as `Point` by default
        return payload_geoproperty("Point")

    elif metadata[1] == "Relationship":
        # TODO only able to return Relationship in string, not in array or object
        return payload_relationship(prop, dataModel)

    elif prop in ["anyOf", "allOf", "oneOf"]:
        # take the first one in the structure to generate, consider as the same level
        return parse_property2ngsild_example({"item": fullPayload[prop][0]}, dataModel, level)

    elif "type" in keys:  # generate property according to its type

        if payload["type"] == "number":
            # treating number with/without maximum/minimum
            argsNumber = (4,)
            if "minimum" in payload:
                argsNumber = argsNumber + (float(payload["minimum"]),)
            else:
                argsNumber = argsNumber + (0,)
            if "maximum" in payload:
                argsNumber = argsNumber + (float(payload["maximum"]),)
            else:
                argsNumber = argsNumber + (max(1000, argsNumber[1] + 1000),)
            return payload_number(argsNumber)

        elif payload["type"] == "integer":
            # treating integer with/without maximum/minimum
            argsNumber = ()
            if "minimum" in payload:
                argsNumber = argsNumber + (int(payload["minimum"]),)
            else:
                argsNumber = argsNumber + (0,)
            if "maximum" in payload:
                argsNumber = argsNumber + (int(payload["maximum"]),)
            else:
                argsNumber = argsNumber + (int(argsNumber[0] + 1000),)
            return payload_integer(argsNumber)

        elif payload["type"] == "string" and "format" in keys:
            # treat string with specific format defined
            if payload["format"] == "date-time":
                return payload_datetime()
            elif payload["format"] == "date":
                return payload_date()
            elif payload["format"] == "time":
                return payload_time()
            elif payload["format"] == "uri":
                return payload_uri(prop, dataModel)
            elif (payload["format"] == "idn-email") or (payload["format"] == "email"):
                return payload_email(dataModel)
            elif payload["format"] == "text":
                return payload_string(1)
            else:  # some format we couldn't generate for now
                return {"type": "Property", "value": "This is a certain format we couldn't generate for now."}

        elif payload["type"] == "string" and "enum" in keys:
            # treat string with enum defined, take only one
            return payload_enum(payload["enum"], 1)

        elif payload["type"] == "string":
            # treat normal string with minLength, maxLength, and pattern
            if "minLength" in payload:
                argsMin = int(payload["minLength"])
            else:
                argsMin = 0
            if "maxLength" in payload:
                argsMax = int(payload["maxLength"])
            else:
                argsMax = 200
            if "pattern" in payload:
                argsPattern = payload["pattern"]
            else:
                argsPattern = ""
            return payload_string(1, argsMin, argsMax, argsPattern)

        elif payload["type"] == "boolean":
            # treat boolean
            return payload_boolean()

        elif payload["type"] == "array":
            # treat array with minItems/maxItems/uniqueItems
            if "minItems" in payload:
                arrayItems = payload["minItems"]
            else:
                arrayItems = 2
            if "maxItems" in payload:
                arrayItems = int((arrayItems + int(payload["maxItems"])) / 2)
            else:
                arrayItems = arrayItems
            if "uniqueItems" in payload:
                uniqueItems = payload["uniqueItems"]
            else:
                uniqueItems = False  # False by default
            arrayPayload = payload["items"].copy()
            if "required" in arrayPayload:
                del arrayPayload["required"]

            valuesArray = []
            idx = 0
            while len(valuesArray) < arrayItems:
                # naive while-loop deals with uniqueItems if it is true
                tmp_value = parse_property2ngsild_example({"items": arrayPayload}, dataModel, level + 1)["value"]
                if uniqueItems and tmp_value in valuesArray:
                    continue
                else:
                    valuesArray.append(tmp_value)

                idx += 1
                if idx >= arrayItems * 5:  # max number of trying
                    # there may be the schema error
                    print("Scheme Error!")
                    break

            return {"type": "Property", "value": valuesArray}

        elif payload["type"] == "object":
            # treat object
            valuesObject = {}
            objectPayload = payload["properties"]
            if "required" in objectPayload:  # ignore required
                del objectPayload["required"]

            for p in objectPayload:  # generate properties for the object
                # only take the `value` to simplify the output, `object` if it's a Relationship
                objectOutput = parse_property2ngsild_example({p: objectPayload[p]}, dataModel, level + 1)
                if objectOutput["type"] == "Relationship":
                    valuesObject[p] = objectOutput["object"]
                else:
                    valuesObject[p] = objectOutput["value"]

            return {"type": "Property", "value": valuesObject}

    elif "oneOf" in keys:
        return parse_property2ngsild_example({prop: fullPayload[prop]["oneOf"][0]}, dataModel, level)

    elif "anyOf" in keys:
        return parse_property2ngsild_example({prop: fullPayload[prop]["anyOf"][0]}, dataModel, level)

    else:
        # other cases
        return {prop: pendingToImplement, "value": fake_uri(prop, dataModel)}

#########################################
#   Data model check related
#########################################

def merge_duplicate_attributes(dict_a: dict, dict_b: dict) -> dict:
    """
    Merge two dictionaries where the values are lists, combining duplicate keys.

    Parameters:
        dict_a (dict): The first dictionary to be merged.
        dict_b (dict): The second dictionary to be merged.

    Returns:
        dict: A new dictionary containing merged values for common keys.

    Example:
    >>> dict_a = {"1": [], "2": ["aa", "bb"]}
    >>> dict_b = {"1": [], "2": ["cc"], "3": ["dd"]}
    >>> merge_duplicate_attributes(dict_a, dict_b)
    {"1": [], "2": ["aa", "bb", "cc"], "3": ["dd"]}
    """
    for key, values in dict_b.items():
        # If the key exists in dict_a, extend its list with values from dict_b
        # If the key doesn't exist, create a new entry with values from dict_b
        dict_a[key] = dict_a.get(key, []) + values
    return dict_a

def parse_payload_v2(schemaPayload, level: int = 1) -> Tuple[Dict, Dict]:
    """
    Parse the given JSON schema payload recursively.

    Parameters:
        schemaPayload: The JSON schema payload to be parsed. could be dict or list
        level (int): The recursion level.

    Returns:
        tuple[dict, dict]: A tuple containing the parsed output and attributes.
    """
    output = {}
    attributes = {level: []}

    def process_allOf_oneOf_anyOf(ofSubSchema):
        nonlocal attributes, output
        for index in range(len(ofSubSchema)):
            partialOutput, partialAttr = parse_payload_v2(ofSubSchema[index], level + 1)
            output = dict(output,
                          **partialOutput["properties"] if "properties" in ofSubSchema[index] else partialOutput)
            attributes = merge_duplicate_attributes(attributes, partialAttr)

    if level == 1:  # first level starts always with "allOf", "properties", "anyOf", "oneOf"

        keys_to_check_level_one = ["allOf", "anyOf", "oneOf", "properties"]
        for key in keys_to_check_level_one:
            if key in schemaPayload:

                # if key is properties, then the value of the key is the sub-structure that needs to be extracted
                # same for "allOf", "anyOf", "oneOf", since their value type is always a list, so process the list first
                # Example
                # {"properties": {...}}
                # {"allOf": [{...}, {...}]}
                if key == "properties":
                    output, partialAttr = parse_payload_v2(schemaPayload[key], level + 1)
                    attributes = merge_duplicate_attributes(attributes, partialAttr)
                else:
                    process_allOf_oneOf_anyOf(schemaPayload[key])

    elif level < 8:  # level from 2 to 8 will be processed as below, in case of the infinite loop or too massive schema

        # if schema is a list, then process the list first, take each element as a sub-schema
        # Example: [{...}, {...}, ...]
        if isinstance(schemaPayload, list):
            for index in range(len(schemaPayload)):
                partialOutput, partialAttr = parse_payload_v2(schemaPayload[index], level + 1)
                output = dict(output, **partialOutput)
                attributes = merge_duplicate_attributes(attributes, partialAttr)

        # if schema is a dictionary, then go inside directly
        # Example: {"allOf": [{...}, ...], "properties": {...}, "description": "xxxx", ...}
        if isinstance(schemaPayload, dict):

            for subschema in schemaPayload:  # run through each key

                if subschema in ["allOf", "anyOf", "oneOf"]:
                    output[subschema] = []
                    for index in range(len(schemaPayload[subschema])):
                        partialOutput, partialAttr = parse_payload_v2(schemaPayload[subschema][index], level + 1)
                        output[subschema].append(
                            partialOutput["properties"] if "properties" in schemaPayload[subschema][
                                index] else partialOutput)
                        attributes = merge_duplicate_attributes(attributes, partialAttr)

                # Extract "prop1", "prop2" as the example given below
                # Example: {"properties": {"prop1": {...}, "prop2": {...}}}
                elif subschema == "properties":

                    output[subschema] = {}
                    for prop in schemaPayload["properties"]:

                        try:
                            output[subschema][prop]
                        except:
                            output[subschema][prop] = {}
                            attributes[level].append(prop)  # get the list of extracted properties per level

                        # process the specific sub-structure of a found property
                        for item in list(schemaPayload["properties"][prop]):

                            # get everything in allOf, anyOf, oneOf under a property
                            if item in ["allOf", "anyOf", "oneOf"]:
                                output[subschema][prop][item] = []
                                for index in range(len(schemaPayload[subschema][prop][item])):
                                    partialOutput, partialAttr = parse_payload_v2(
                                        schemaPayload[subschema][prop][item][index], level + 1)
                                    output[subschema][prop][item].append(partialOutput)
                                    attributes = merge_duplicate_attributes(attributes, partialAttr)

                            # parse the description, return a list [description, NGSI type, model name, unit name, enum, privacy]
                            elif item == "description":
                                x_ngsi, description = parse_schema_description(schemaPayload[subschema][prop][item])
                                output[subschema][prop][item] = description
                                if x_ngsi:
                                    output[subschema][prop]["x-ngsi"] = x_ngsi

                            # parse the sub-structure if equals to items or properties
                            elif item in ["items", "properties"]:
                                output[subschema][prop][item], partialAttr = parse_payload_v2(
                                    schemaPayload[subschema][prop][item], level + 1)
                                attributes = merge_duplicate_attributes(attributes, partialAttr)

                            # change the type integer to number
                            elif item == "type":
                                output[subschema][prop][item] = "number" if schemaPayload[subschema][prop][
                                                                                item] == "integer" else \
                                schemaPayload[subschema][prop][item]

                            # remain the rest part the same
                            else:
                                output[subschema][prop][item] = schemaPayload[subschema][prop][item]

                # if equals to description, parse it
                elif subschema == "description":
                    x_ngsi, description = parse_schema_description(schemaPayload[subschema])
                    output[subschema] = description
                    if x_ngsi:
                        output["x-ngsi"] = x_ngsi

                # for the subschema that is a dictionary, parse the inside
                elif isinstance(schemaPayload[subschema], dict):
                    attributes[level].append(subschema)
                    output[subschema], partialAttr = parse_payload_v2(schemaPayload[subschema], level + 1)
                    attributes = merge_duplicate_attributes(attributes, partialAttr)

                # for the rest, remain the same
                else:
                    output[subschema] = schemaPayload[subschema]

    else:  # return None when parsing over 8 levels
        return None, None

    return output, attributes

def parse_yamlDict(yamlDict, datamodelRepoUrl: str, level: int = 1) -> dict:
    """
    Check the properties of descriptions, NGSI types, etc

    Parameters:
    - yamlDict (dict or list): The YAML dictionary to parse.
    - datamodelRepoUrl (str): The URL of the data model repository.
    - level (int, optional): The nesting level in the YAML structure. Defaults to 1.

    Returns:
    dict: A dictionary containing the extracted properties following the check results.

    Example:
        yamlDict = {"prop1": {"description": xxx, "type": "string", "x-ngsi": {"type": "Property"}, ...}, "prop2": ...}
        parse_yamlDict(yamlDict, datamodelRepoUrl)
    {
        "prop1": {
            "x-ngsi": true,
            "x-ngsi_text": "ok to Property",
            "documented": true,
            "text": "xxx"
        },
        "prop2": ...
    }
    """
    output = {}

    # if the yamlDict is a list, then parse it individually
    if isinstance(yamlDict, list):
        for item in yamlDict:
            partialoutput = parse_yamlDict(item, datamodelRepoUrl, level + 1)
            output = dict(output, **partialoutput)

    else:

        for prop in yamlDict:

            if prop == "id": continue

            # Example: [{...}, {...}, ...]
            if isinstance(yamlDict[prop], list) and len(yamlDict[prop]) > 1 and isinstance(yamlDict[prop][0], dict):
                for item in yamlDict[prop]:
                    partialoutput = parse_yamlDict(item, datamodelRepoUrl, level + 1)
                    output = dict(output, **partialoutput)

            # Example: {"description": xxx, "properties": xxx, "x-ngsi": xxx, ...}
            if isinstance(yamlDict[prop], dict):

                # if prop equals to any value of "properties", "allOf", "oneOf", "anyOf", "items"
                # means the value of prop has deeper structure, like dictionary or list
                # so need to parse it one level deeper
                if prop in ["properties", "allOf", "oneOf", "anyOf", "items"]:
                    partialoutput = parse_yamlDict(yamlDict[prop], datamodelRepoUrl, level + 1)
                    output = dict(output, **partialoutput)
                    continue

                # Process the substructures
                # Get the keys inside
                # Examples: {"description": xxx, "properties": xxx, "x-ngsi": xxx, ...}
                # --> ["description", "properties", "x-ngsi", ...]
                for propKey in yamlDict[prop].keys():
                    if propKey in ["properties", "anyOf", "allOf", "oneOf", "items"]:
                        partialoutput = parse_yamlDict(yamlDict[prop][propKey], datamodelRepoUrl, level + 1)
                        output = dict(output, **partialoutput)

                # prop is a real property except it equals to `x-ngsi` and inside of exceptions, exceptions = ["coordinates", "bbox", "type"]
                # "x-ngsi": {"type": "Property",  "model": xxx}
                # exceptions: no descriptions and no NGSI type claimed
                if prop != "x-ngsi" and not prop in exceptions:

                    try:
                        # Process the NGSI type, have to be one of the three ["Property", "Relationship", "GeoProperty"]
                        propertyType = yamlDict[prop]["x-ngsi"]["type"]

                        if propertyType in propertyTypes:
                            output[prop] = {}
                            output[prop]["x-ngsi"] = True
                            output[prop]["x-ngsi_text"] = "ok to " + str(propertyType)
                        else:
                            output[prop]["x-ngsi"] = False
                            output[prop][
                                "x-ngsi_text"] = "Wrong NGSI type of " + propertyType + " in the description of the property on level " + str(
                                level)

                    except:
                        output[prop] = {}
                        output[prop]["x-ngsi"] = False
                        output[prop]["x-ngsi_text"] = "Missing NGSI type of " + str(
                            propertyTypes) + " in the description of the property on level " + str(level)

                    try:
                        # checking the pure description whether is well-documented
                        description = yamlDict[prop]["description"]

                        if len(description) > 15:  # if the description is longer than 15 char then it is well-documented
                            output[prop]["documented"] = True
                            output[prop]["text"] = description
                        else:
                            output[prop]["documented"] = False
                            output[prop]["text"] = incompleteDescription

                    except:
                        output[prop]["documented"] = False
                        output[prop]["text"] = withoutDescription

                    if prop == "type" and level == 1:

                        try:
                            # Type property matches data model name
                            propertyType = yamlDict[prop]["enum"]

                            if propertyType[0] == extract_datamodel_from_raw_url(datamodelRepoUrl):
                                output[prop]["type_specific"] = True
                                output[prop][
                                    "type_specific_text"] = "Type property matches to data model name on level " + str(
                                    level)
                            else:
                                output[prop]["type_specific"] = False
                                output[prop][
                                    "type_specific_text"] = "Type property doesn't match to data model name on level " + str(
                                    level)

                        except:
                            output[prop]["type_specific"] = False
                            output[prop]["type_specific_text"] = "Missing Type property"

    return output

def is_metadata_properly_reported(output: dict, schemaDict: dict) -> dict:
    """
    Check if metadata (derivedFrom and license) is properly reported in the output based on the given JSON schema.

    Parameters:
    - output (dict): The output dictionary to report metadata warnings.
    - schemaDict (dict): The schema dictionary containing information about the schema.

    Returns:
    - dict: The updated output dictionary with metadata warnings.

    Example:
    >>> schema = {"derivedFrom": "https://example.com/base_schema", "license": "https://example.com/license"}
    >>> is_metadata_properly_reported({}, schema)
    {'metadata': {'derivedFrom': {'warning': 'derivedFrom is not a valid url'},
                  'license': {'warning': 'License is not a valid url. It should be a link to the license document'}}}

    """
    if "metadata" not in output:
        output["metadata"] = {}

    metadata_fields = ["derivedFrom", "license"]

    for field in metadata_fields:
        try:
            # if field not in output["metadata"]:
            #     output["metadata"][field] = {}

            if field in schemaDict:
                value = schemaDict[field]

                if value:
                    if not checkers.is_url(value):
                        # check url format
                        output["metadata"][field] = {"warning": f"{field} is not a valid URL"}
                    elif not is_url_existed(value)[0]:
                        # check the url reachability
                        output["metadata"][field] = {"warning": f"{field} URL is not reachable"}
                else:
                    output["metadata"][field] = {"warning": f"{field} is empty"}
            else:
                output["metadata"][field] = {"warning": f"Missing {field} clause, include {field} = '' in the header"}
        except Exception as e:
            output["metadata"][field] = {"warning": f"not possible to check {field} clause, {str(e)}"}

    return output

def is_metadata_existed(output: dict, jsonDict: dict, schemaUrl: str, message: str = "", checkall: bool = True,
                        checklist: list = None) -> dict:
    """
    Check if metadata (e.g., $schema, $id, title, description, modelTags, schemaVersion, required) is properly reported
    in the output based on the given schema dictionary.

    Parameters:
    - output (dict): The output dictionary to report metadata warnings.
    - jsonDict (dict): The schema dictionary containing information about the schema.
    - schemaUrl (str): The URL of the schema.
    TODO: extend function with three parameters below
    - message (str): Custom message.
    - checkall (bool): If True, check all metadata; if False, use the checklist.
    - checklist (list): List of metadata items to check.

    Returns:
    - dict: The updated output dictionary with metadata warnings.

    Example:
    >>> schema = {"$schema": "http://json-schema.org/schema#", "$id": "https://example.com/schema", "title": "Example Schema"}
    >>> is_metadata_existed({}, schema, "https://example.com/schema")
    {'metadata': {'$schema': {'warning': '$schema should be "http://json-schema.org/schema#" by default'},
                  '$id': {'warning': '$id doesn't match, please check it again'},
                  'title': {'warning': 'Title too short'},
                  'description': {'warning': 'Description is empty'},
                  'modelTags': {'warning': 'Missing modelTags clause, include modelTags = \'\' in the header'},
                  'schemaVersion': {'warning': 'Schema version format wrong. Right is x.x.x'},
                  'required': {'warning': 'Too many required attributes, consider its reduction to less than 5 preferably just id and type'}}}

    """
    if "metadata" not in output:
        output["metadata"] = {}

    metadata_fields = {
        "$schema": {"type": "string", "value": "http://json-schema.org/schema#"},
        "$id": {"type": "string", "value": schemaUrl},
        "title": {"type": "string", "min_length": 15},
        "description": {"type": "string", "min_length": 34},
        "modelTags": {"type": "string"},
        "$schemaVersion": {"type": "string", "pattern": "^\d{1,3}.\d{1,3}.\d{1,3}$"},
        "required": {"type": "list", "default": ["id", "type"], "max_length": 4, "mandatory_fields": ["id", "type"]}
    }

    for field, constraints in metadata_fields.items():
        try:
            # if field not in output["metadata"]:
            #     output["metadata"][field] = {}

            if field in jsonDict:
                value = jsonDict[field]

                # match with the type
                if constraints["type"] == "string":
                    if not isinstance(value, str):
                        output["metadata"][field] = {"warning": f"{field} is not a string"}
                elif constraints["type"] == "list":
                    if not isinstance(value, list):
                        output["metadata"][field] = {"warning": f"{field} is not a list"}

                # match with the value
                if "value" in constraints:
                    if value != constraints:
                        if field == "$schema":
                            output["metadata"][field] = {"warning": f"{field} should be {constraints} by default"}
                        elif field == "$id" and (KEYWORDS_FOR_CERTAIN_CHECK in constraints):
                            output["metadata"][field] = {
                                "warning": f"{field} doesn't match, please check it again or ignore this message if it's an unpublished datamodel"}

                # min_length
                if "min_length" in constraints and len(value) < constraints["min_length"]:
                    output["metadata"][field] = {"warning": f"{field} is too short"}

                # max_length
                if "max_length" in constraints and len(value) > constraints["max_length"]:
                    output["metadata"][field] = {"warning": f"{field} is too long"}

                # pattern
                if "pattern" in constraints and re.search(constraints["pattern"], value) is None:
                    output["metadata"][field] = {"warning": f"{field} format wrong. Right is x.x.x. Default = 0.0.1"}

                # mandatory_field
                if "mandatory_fields" in constraints:
                    for mf in constraints["mandatory_fields"]:
                        if mf not in value:
                            output["metadata"][field] = {
                                "warning": f"{field} should include {', '.join(constraints['mandatory_fields'])}"}
            else:
                output["metadata"][field] = {"warning": f"Missing {field} clause, include {field} = '' in the header"}
        except Exception as e:
            output["metadata"][field] = {"warning": f"not possible to check {field} clause, {str(e)}"}

    return output

def schema_output_sum(output: dict) -> dict:
    """
    Summarize the output of a schema validation process, categorizing properties and metadata, especially for the failed cases.

    Parameters:
    - output (dict): The output dictionary from a schema validation process.

    Returns:
    - dict: A summary dictionary categorizing properties and metadata.

    Example:
        validation_output = {...}  # An example output from a schema validation process
        schema_output_summary(validation_output)
    {
        'well documented': [...],
        'already used': [...],
        'newly available': [...],
        'Metadata': [...],
        'Failed': {
            'Failed reason 1': [...],
            'Failed reason 2': [...],
            ...
        },
        ...
    }
    """
    documentationStatusOfProperties = output['documentationStatusOfProperties']
    alreadyUsedProperties = output['alreadyUsedProperties']
    availableProperties = output['availableProperties']
    metadata = output['metadata']

    results = {}
    results = {key: [] for key in CHECKED_PROPERTY_CASES}
    results['Failed'] = {}

    for prop, value in documentationStatusOfProperties.items():

        if value['documented'] & value['x-ngsi']:
            # Only if the `documented` and NGSI type are True
            results['well documented'].append(prop)

        else:
            if value['x-ngsi'] is False:
                # If NGSI type is False
                if value['x-ngsi_text'] not in results['Failed'].keys():
                    results['Failed'][value['x-ngsi_text']] = []
                results['Failed'][value['x-ngsi_text']].append(prop)

            if value['documented'] is False:
                # If documented is False
                if value['text'] not in results['Failed'].keys():
                    results['Failed'][value['text']] = []
                results['Failed'][value['text']].append(prop)

        if (prop == "type") and (value["type_specific"] is False):
            # If the property is `type`, then it has to be in the same as datamodel
            # If it is different, then `type_specific` is False
            results['Failed'][value["type_specific_text"]] = []
            results['Failed'][value["type_specific_text"]].append(prop)

        if 'duplicated_prop' in value:
            # Collect duplicated properties
            try:
                results['Failed'][value['duplicated_prop_text']].append(prop)
            except:
                results['Failed'][value['duplicated_prop_text']] = []
                results['Failed'][value['duplicated_prop_text']].append(prop)

    for prop in alreadyUsedProperties:
        results['already used'].append(list(prop.keys())[0])

    for prop in availableProperties:
        results['newly available'].append(list(prop.keys())[0])

    for prop, value in metadata.items():
        results['Metadata'].append(value['warning'])

    return results

def message_after_check_schema(results: dict) -> str:
    """
    Generate a summary message based on the results of a schema validation check.

    Parameters:
    - results (dict): The summarized results from schema validation.

    Returns:
    - str: A message providing information about different property categories and metadata warnings.

    Example:
    >>> validation_results = {...}  # An example of summarized validation results
    >>> message_after_check_schema(validation_results)
    '
    These properties are well documented properties:

        dateCreated, dateModified, source, name, ...

    These properties are already used properties:

        openingHoursSpecification, startDate, ...

    No big issue with the named properties in general.

    Some warnings related to metadata:

    ...
    '
    """
    message = ""

    for key in CHECKED_PROPERTY_CASES[:-2]:
        if results[key]:
            message += f"\nThese properties are {key} properties:\n    {newline.join(results[key])}"

    # Include messages for properties with failures
    if results[CHECKED_PROPERTY_CASES[-1]]:
        message += f"""
However, We highly suggest you fix these properties:
    {newline.join([" - " + text + newline + f"{', '.join(pps)}" for text, pps in results['Failed'].items()])}
        """
    else:
        message += "\nNo major issues with the named properties in general."

    # Include messages for metadata warnings
    if results[CHECKED_PROPERTY_CASES[-2]]:
        message += f"\nSome warnings related to metadata:\n    {newline.join([' - ' + text for text in results['Metadata']])}"
    else:
        message += "\nNo warnings with metadata."

    return message

def convert_to_raw_github_url(input_url):
    match = re.match(github_url_pattern, input_url)
    
    if match:
        owner = match.group(1)
        repository = match.group(2)
        branch = match.group(3)
        file_path = match.group(4)
        raw_github_url = f"https://raw.githubusercontent.com/{owner}/{repository}/{branch}/{file_path}"
        return raw_github_url
    else:
        print("Invalid GitHub URL format")


import datetime

import re

import json
import os
import sys
import urllib.request

import jsonschema
import pytz
import requests

from jsonschema import validate, Draft202012Validator

import urllib.request

path = __file__

# Find the index of the last occurrence of /
index = path.rfind(os.sep)

# Extract the right part of the string up until the last /
left_part = path[:index]
# lookup install path for - model-assets

from pysmartdatamodels.utils.common_utils import (
    create_context,
    extract_datamodel_from_raw_url,
    extract_subject_from_raw_url,
    generate_random_string,
    is_metadata_properly_reported,
    is_metadata_existed,
    is_url_existed,
    message_after_check_schema,
    normalized2keyvalues_v2,
    open_jsonref,
    open_yaml,
    parse_payload_v2,
    parse_property2ngsild_example,
    parse_yamlDict,
    schema_output_sum,
)

official_list_file_name = left_part + "/model-assets/official_list_data_models.json"
ddbb_attributes_file = left_part + "/model-assets/smartdatamodels.json"
metadata_file = left_part + "/model-assets/datamodels_metadata.json"

# data models that are not able to generate examples from the schema due to certain issue
pending_datamodels = ["SeaportFacilities", "OffStreetParking", "OnStreetParking", "ParkingGroup", "WeatherAlert",
                      "Museum", "MosquitoDensity", "Alert", "MediaEvent", "DataResourceFrictionlessData",
                      "TableSchemaFrictionlessData", "free_bike_status", "geofencing_zones", "station_information",
                      "station_status", "system_alerts", "Presentation", "DigitalInnovationHub"]
returnPendingMessage = "====== Examples are unable to be generated for the model ====== \nThe current schema may includes open attributes that need further definition, like subproperties inside of object structure, or other inner problems, like infinite recursion and unknown format."

# Regular expression pattern to extract owner, repository, branch, and file path
github_url_pattern = r"https://github\.com/([^/]+)/([^/]+)/blob/([^/]+)/(.+)"


# 1
def load_all_datamodels():
    """Returns a dict with all data models with this object structure
        - repoName: The name of the subject
        - repoLink: the link to the repository of the subject
        - dataModels: An array with all the datamodels of this subject
        - domains: an array to the domains that this subject belongs to
        Parameters:

        Returns:
           array of objects with the description of the subject
        """
    import json

    output = []
    # Opens the file with the list of data models
    with open(official_list_file_name, "r", encoding='utf-8') as list_of_datamodels_pointer:
        output = json.load(list_of_datamodels_pointer)["officialList"]
    return output


# 2
def load_all_attributes():
    """Returns an array of objects describing every attribute in the data models
        - _id: identifier of the item
        - property: the name of the attribute
        - dataModel: the data model this attribute is present
        - repoName: the subject this data model belongs to
        - description: the description of the attribute
        - typeNGSI: Whether it is a Property, GeoProperty, or Relationship
        - modelTags: inherited from the data model tags
        - license: link to the license for the data model
        - schemaVersion: version of the data model
        - type: data type
        - model: when available the reference model for the attribute
        - units: when available the recommended units for the attribute
        - format: either date, or time, or date-time, or URI, etc the format of the attribute
    Parameters:

    Returns:
        array of objects with the description of the subject
    """

    output = []
    # Opens the file with the list of data models
    with open(ddbb_attributes_file, "r", encoding='utf-8') as ddbb_attributes_file_pointer:
        output = json.load(ddbb_attributes_file_pointer)
    return output


# 3
def list_all_datamodels():
    """List the names of the entities defined in the data models.
    Parameters:

    Returns:
       array of strings: data models' names
    """

    output = []
    # Opens the file with the list of data models
    with open(official_list_file_name, "r", encoding='utf-8') as list_of_datamodels_pointer:
        # Loads it in the dictionary
        datamodelsdict = json.load(list_of_datamodels_pointer)["officialList"]
        # Stores it in the output variable
        for item in datamodelsdict:
            for datamodel in item["dataModels"]:
                output.append(datamodel)

    return output


# 4
def list_all_subjects():
    """List the names of the subjects (groups of data models). The subject's names define repositories with the name dataModel.subject at the root of the https://smart-data-models.github.com site
    Parameters:

    Returns:
       array of strings: subjects' names
    """

    output = []

    # Opens the file with the list of data models
    with open(official_list_file_name, "r", encoding='utf-8') as list_of_datamodels_pointer:
        datamodelsdict = json.load(list_of_datamodels_pointer)["officialList"]
        for item in datamodelsdict:
            output.append(item["repoName"])

    return output


# 5
def datamodels_subject(subject: str):
    """List the names of the data models defined in the subject.
    Parameters:
        subject: name of the subject

    Returns:
        if subject is found
            array of strings: data models' names belonging to the subject
        if subject is not found
            False
    """

    output = []
    done = False

    with open(official_list_file_name, "r", encoding='utf-8') as list_of_datamodels_pointer:
        datamodelsdict = json.load(list_of_datamodels_pointer)["officialList"]
        for item in datamodelsdict:
            if "repoName" and "dataModels" in item:
                if item["repoName"] == subject:
                    output = item["dataModels"]
                    done = True
    if not done:
        output = False

    return output


# 6
def description_attribute(subject, datamodel, attribute):
    """List the description of an attribute belonging to a subject and data model.
    Parameters:
        subject: name of the subject
        datamodel: name of the data model
        attribute: name of the attribute

    Returns:
        if subject, datamodel and attribute are found
            string: attribute's description
        if any of the input parameters is not found
            False
    """

    output = []
    done = False

    # Access the full database of attributes and stores is in a dictionary
    with open(ddbb_attributes_file, "r", encoding='utf-8') as ddbb_attributes_pointer:
        datamodelsdict = json.load(ddbb_attributes_pointer)

    # Looks for the attribute in the dictionary
    for item in datamodelsdict:
        try:
            if (
                    (item["repoName"] == subject)
                    and (item["dataModel"] == datamodel)
                    and (item["property"] == attribute)
            ):
                output = item["description"]
                done = True
        except:
            nada = 0
    else:
        if not done:
            output = False

    return output


# 7
def datatype_attribute(subject, datamodel, attribute):
    """List the data type of an attribute belonging to a subject and data model.
    Parameters:
        subject: name of the subject
        datamodel: name of the data model
        attribute: name of the attribute

    Returns:
        if subject, datamodel and attribute are found
            string: attribute's data type
        if any of the input parameters is not found
            False
    """

    output = []
    done = False

    # Access the full database of attributes and stores is in a dictionary
    with open(ddbb_attributes_file, "r", encoding='utf-8') as ddbb_attributes_pointer:
        datamodelsdict = json.load(ddbb_attributes_pointer)

    # Looks for the attribute in the dictionary
    for item in datamodelsdict:
        # print(item)
        try:
            if (
                    (item["repoName"] == subject)
                    and (item["dataModel"] == datamodel)
                    and (item["property"] == attribute)
            ):
                if "type" in item:
                    output = item["type"]
                    done = True
                else:
                    output = False
                    done = True
        except:
            done = False
    else:
        if not done:
            output = False

    return output


# 8
def model_attribute(subject, datamodel, attribute):
    """List the model of an attribute (when available) belonging to a subject and data model.
    Parameters:
        subject: name of the subject
        datamodel: name of the data model
        attribute: name of the attribute

    Returns:
        if subject, datamodel and attribute are found
            string: attribute model's URL
        if any of the input parameters is not found or there is not a model
            False
    """

    output = []
    done = False

    # Access the full database of attributes and stores is in a dictionary
    with open(ddbb_attributes_file, "r", encoding='utf-8') as ddbb_attributes_pointer:
        datamodelsdict = json.load(ddbb_attributes_pointer)

    # Looks for the attribute in the dictionary
    for item in datamodelsdict:
        # print(item)
        try:
            if (
                    (item["repoName"] == subject)
                    and (item["dataModel"] == datamodel)
                    and (item["property"] == attribute)
            ):
                if "model" in item:
                    output = item["model"]
                    done = True
                else:
                    output = False
                    done = True
        except:
            nada = 0
    else:
        if not done:
            output = False

    return output


# 9
def units_attribute(subject, datamodel, attribute):
    """List the recommended units of an attribute belonging to a subject and data model.
    Parameters:
        subject: name of the subject
        datamodel: name of the data model
        attribute: name of the attribute

    Returns:
        if subject, datamodel and attribute are found
            string: acronym/text of the recommended units
        if any of the input parameters is not found or there are not recommended units
            False
    """

    output = []
    done = False

    # Access the full database of attributes and stores is in a dictionary
    with open(ddbb_attributes_file, "r", encoding='utf-8') as ddbb_attributes_pointer:
        datamodelsdict = json.load(ddbb_attributes_pointer)

    # Looks for the attribute in the dictionary
    for item in datamodelsdict:
        # print(item)
        try:
            if (
                    (item["repoName"] == subject)
                    and (item["dataModel"] == datamodel)
                    and (item["property"] == attribute)
            ):
                if "units" in item:
                    output = item["units"]
                    done = True
                else:
                    output = False
                    done = True
        except:
            nada = 0
    else:
        if not done:
            output = False

    return output


# 10
def attributes_datamodel(subject, datamodel):
    """List the attributes of a data model (currently only first level ones) .
    Parameters:
        subject: name of the subject
        datamodel: name of the data model

    Returns:
        if subject and datamodel  are found
            array: attribute's names
        if any of the input parameters is not found
            False
    """

    output = []
    done = False

    # Access the full database of attributes and stores is in a dictionary
    with open(ddbb_attributes_file, "r", encoding='utf-8') as ddbb_attributes_pointer:
        datamodelsdict = json.load(ddbb_attributes_pointer)

    # Looks for the attribute in the dictionary
    for item in datamodelsdict:
        # print(item)
        try:
            if (item["repoName"] == subject) and (item["dataModel"] == datamodel):
                if "property" in item:
                    output.append(item["property"])
                    done = True
        except:
            nada = 0
    else:
        if len(output) < 1:
            output = False
        if not done:
            output = False

    return output


# 11
def ngsi_datatype_attribute(subject, datamodel, attribute):
    """List the NGSI data type of an attribute (Property, Relationship or GeoProperty) belonging to a subject and data model.
    Parameters:
        subject: name of the subject
        datamodel: name of the data model
        attribute: name of the attribute

    Returns:
        if subject, datamodel and attribute are found
            string: NGSI data type
        if any of the input parameters is not found
            False
    """

    output = []
    done = False

    # Access the full database of attributes and stores is in a dictionary
    with open(ddbb_attributes_file, "r", encoding='utf-8') as ddbb_attributes_pointer:
        datamodelsdict = json.load(ddbb_attributes_pointer)

    # Looks for the attribute in the dictionary
    for item in datamodelsdict:
        # print(item)
        try:
            if (
                    (item["repoName"] == subject)
                    and (item["dataModel"] == datamodel)
                    and (item["property"] == attribute)
            ):
                if "typeNGSI" in item:
                    output = item["typeNGSI"]
                    done = True
        except:
            nada = 0

    if done:
        return output
    else:
        return False


# 12
def validate_data_model_schema(schema_url):
    """Validates a json schema defining a data model.
    Parameters:
        schema_url: url of the schema (public available). (i.e. raw version of a github repo https://raw.githubusercontent.com/smart-data-models/dataModel.Aeronautics/master/AircraftModel/schema.json

    Returns:
        object with four elements:
        - documentationStatusofProperties: For each first level attribute lists if the attribute is documented and includes the description (when available). Also the NGSI type if is set and which one is described.
          Example:
            "dateCreated":
                  {
                  "x-ngsi": true,
                  "x-ngsi_text": "ok to Property",
                  "documented": true,
                  "text": "This will usually be allocated by the storage platform. Entity creation timestamp"
                  },
        - schemaDiagnose: It counts the attributes with right descriptions and those which don't.
        - alreadyUsedProperties: It identifies attributes that have already been used in other data models and includes their definition
        - availableProperties: Identifies those attributes which are not already included in any other data model
    """

    def is_property_already_existed(output: dict, yamlDict: dict) -> dict:
        """Sorts out the properties already existed in the database

        Parameters:
            output (dict): The output dictionary to report properites.
            yamlDict (dict): The yaml dictionary containing information of properties

        Returns:
            dict: The updated output dictionary with property warnings.
        """
        try:
            commonProperties = ["id", "name", "description", "location", "seeAlso", "dateCreated", "dateModified",
                                "source", "alternateName", "dataProvider", "owner", "address", "areaServed", "type"]
            existing = "alreadyUsedProperties"
            available = "availableProperties"

            output[existing] = []
            output[available] = []

            for key in yamlDict:
                if key in commonProperties:
                    continue
                lowKey = key.lower()

                with open(ddbb_attributes_file, "r", encoding='utf-8') as ddbb_attributes_pointer:
                    datamodelsdict = json.load(ddbb_attributes_pointer)

                results = []
                for item in datamodelsdict:
                    try:
                        if re.match(lowKey, item['property'], re.IGNORECASE):
                            results.append(item)
                    except:
                        nada = 0

                if len(results) > 0:
                    definitions = []
                    dataModelsList = []
                    types = []
                    for index, item in enumerate(results):
                        dataModelsList.append(str(index + 1) + ".-" + item["dataModel"])
                        if "description" in item or "type" in item:
                            if "description" in item:
                                definitions.append(str(index + 1) + ".-" + item["description"])
                            else:
                                definitions.append(str(index + 1) + ".- missing description")
                            if "type" in item:
                                types.append(str(index + 1) + ".-" + item["type"])
                            else:
                                types.append(str(index + 1) + ".- missing type")
                        else:
                            output[existing].append({"Error": lowKey})
                    output[existing].append({key: "Already used in data models: " + ",".join(
                        dataModelsList) + " with these definitions: " + chr(13).join(
                        definitions) + " and these data types: " + ",".join(types)})
                else:
                    output[available].append({key: "Available"})

        except:
            output[existing].append({"Error": lowKey})

        return output

    # initialize variables for the script
    output = {}  # the json answering the test
    tz = pytz.timezone("Europe/Madrid")
    metaSchema = open_jsonref("https://json-schema.org/draft/2020-12/meta/validation")
    incompleteDescription = "Incomplete description"
    withoutDescription = "No description at all"

    # validate inputs
    existsSchema = is_url_existed(schema_url)

    # url provided is an existing url
    if not existsSchema[0]:
        output["result"] = False
        output["cause"] = "Cannot find the schema at " + schema_url
        output["time"] = str(datetime.datetime.now(tz=tz))
        print(json.dumps(output))
        sys.exit()

    # url is actually a json
    try:
        schemaDict = json.loads(existsSchema[1])
    except ValueError:
        output["result"] = False
        output["cause"] = "Schema " + schema_url + " is not a valid json"
        output["time"] = str(datetime.datetime.now(tz=tz))
        output["parameters"] = {"schema_url: ": schema_url}
        print(json.dumps(output))
        sys.exit()

    # test that it is a valid schema against the metaschema
    try:
        schema = open_jsonref(schema_url)
        if not bool(schema):
            output["result"] = False
            output["cause"] = "json schema returned empty (wrong $ref?)"
            output["time"] = str(datetime.datetime.now(tz=tz))
            output["parameters"] = {"schema_url: ": schema_url}
            print(json.dumps(output))
            sys.exit()

    except:
        output["result"] = False
        output["cause"] = "json schema cannot be fully loaded"
        output["time"] = str(datetime.datetime.now(tz=tz))
        output["parameters"] = {"schema_url": schema_url}
        print(json.dumps(output))
        sys.exit()

    try:
        validate(instance=schema, schema=metaSchema, format_checker=Draft202012Validator.FORMAT_CHECKER)
    except jsonschema.exceptions.ValidationError as err:
        # print(err)
        output["result"] = False
        output["cause"] = "schema does not validate as a json schema"
        output["time"] = str(datetime.datetime.now(tz=tz))
        output["parameters"] = {"schema_url": schema_url}
        output["errorSchema"] = str(err)
        print(json.dumps(output))
        sys.exit()

    # extract properties' definitions
    # check if they are populated
    documented = "documentationStatusOfProperties"
    try:
        yamlDict, attributes = parse_payload_v2(schema, 1)
    except:
        output["result"] = False
        output["cause"] = "schema cannot be loaded (possibly invalid $ref)"
        output["time"] = str(datetime.datetime.now(tz=tz))
        output["parameters"] = {"schema_url": schema_url}
        print(json.dumps(output))
        sys.exit()

    # check the duplicated attributes
    if len(attributes[2]) != len(set(attributes[2])):

        def find_duplicates(input_list: list) -> list:
            """
            Find and output the duplicated values inside a list.
            Parameters:
                input_list (list): The list to search for duplicates in.
            Returns:
                list: A list containing the duplicated values found in the input list.
            """
            seen = set()
            duplicates = set()

            for item in input_list:
                if item in seen:
                    duplicates.add(item)
                else:
                    seen.add(item)

            return list(duplicates)

        output[
            "cause"] = f"Duplicated attributes (User-defined properties is duplicated with system-defined properties):\n\t{', '.join(find_duplicates(attributes[2]))}"
        output["time"] = str(datetime.datetime.now(tz=tz))
        output["parameters"] = {"schema_url": schema_url}
        print(json.dumps(output))
        sys.exit()

    output[documented] = parse_yamlDict(yamlDict, schema_url, 1)
    allProperties = 0
    documentedProperties = 0
    faultyDescriptionProperties = 0
    notDescribedProperties = 0
    for key in output[documented]:
        allProperties += 1
        if output[documented][key]["documented"]:
            documentedProperties += 1
        elif output[documented][key]["text"] == incompleteDescription:
            faultyDescriptionProperties += 1
        elif output[documented][key]["text"] == withoutDescription:
            notDescribedProperties += 1

    output["schemaDiagnose"] = (
            "This schema has "
            + str(allProperties)
            + " properties. "
            + str(notDescribedProperties)
            + " properties are not described at all and "
            + str(faultyDescriptionProperties)
            + " have descriptions that must be completed. "
            + str(allProperties - faultyDescriptionProperties - notDescribedProperties)
            + " are described but you can review them anyway. "
    )

    # now it checks if these properties already exist in the database
    # TODO
    output = is_property_already_existed(output, yamlDict)

    # now it checks if the metadata is properly reported
    output = is_metadata_properly_reported(output, schemaDict)

    # now it checks if the metadata does exist
    output = is_metadata_existed(output, schemaDict, schema_url, message="schema")

    # make a summary of all the output
    results = schema_output_sum(output)

    # print out the collective message after the schema check
    print(message_after_check_schema(results))

    return output


# 13 print data models attributes
def print_datamodel(subject, datamodel, separator, meta_attributes):
    """print the different elements of the attributes of a data model separated by a given separator.
    Parameters:
        subject: name of the subject
        datamodel: name of the data model
        separator: string between the different elements printed
        meta_attributes: list of different qualifiers of an attribute
             property: the name of the attribute
             type: the data type of the attribute (json schema basic types)
             dataModel: the data model the attribute belongs to
             repoName: the subject the attribute belongs to
             description: the definition of the attribute
             typeNGSI: the NGSI type, Property, Relationship or Geoproperty
             modelTags: the tags assigned to the data model
             format: For those attributes having it the format, i.e. date-time
             units: For those attributes having it the recommended units, i.e. meters
             model: For those attributes having it the reference model, i.e. https://schema.org/Number

    Returns:
        It prints a version of the attributes separated by the separator listing the meta_attributes specified
        A variable with the same strings
        if any of the input parameters is not found it returns false
            False
    """

    output = []
    with open(ddbb_attributes_file, "r", encoding='utf-8') as ddbb_attributes_pointer:
        datamodelsdict = json.load(ddbb_attributes_pointer)

    # available metadata in the list
    validmetadata = [
        "property",
        "type",
        "dataModel",
        "repoName",
        "description",
        "typeNGSI",
        "modelTags",
        "format",
        "units",
        "model",
    ]
    defaultmetadata = ["property", "type", "typeNGSI", "description"]
    newline = chr(13) + chr(10)
    with open(ddbb_attributes_file, "r", encoding='utf-8') as ddbb_attributes_pointer:
        datamodelsdict = json.load(ddbb_attributes_pointer)
    print(datamodelsdict[0])
    selectedattributes = []
    for d in datamodelsdict:
        print(d)
        if "dataModel" in d and "repoName" in d:
            if d["dataModel"] == datamodel and d["repoName"] == subject:
                selectedattributes.append(d)
    # checks that there are any attribute for the data model
    if len(selectedattributes) < 1:
        return False
    else:
        if all([d in validmetadata for d in meta_attributes]):
            listedmetadata = meta_attributes
        else:
            listedmetadata = defaultmetadata
        # first line with the names of the meta data of the attributes
        output = str(separator.join(listedmetadata)) + newline

        # for every attribute in the data model
        for item in selectedattributes:
            print("item:" + str(item))

            try:
                # if all metadata are available for the attribute it is done in one shot
                selectedmeta = [item[d] for d in listedmetadata]
                print("selectedmeta:" + str(selectedmeta))
                row = separator.join(selectedmeta)
                print("row:" + str(row))
            except:
                # if all metadata are not available for the attribute it is done with a loop
                print("error")
                rowitems = []
                for d in listedmetadata:
                    if d in item:
                        rowitems.append(item[d])
                    else:
                        rowitems.append("")
                row = separator.join(rowitems)
            output += row + newline
    print(output)
    return output


# 14
def subject_repolink(subject: str):
    """It returns the direct link to the repository of the subject if it is found and False if not .
    Parameters:
        subject: name of the subject

    Returns:
        if subject is found
            url of the github repository. Example for subject User it returns 'https://github.com/smart-data-models/dataModel.User.git'
        if subject is not found
            False
    """

    # output = []
    done = False

    with open(official_list_file_name, "r", encoding='utf-8') as list_of_datamodels_pointer:
        datamodelsdict = json.load(list_of_datamodels_pointer)["officialList"]
        for item in datamodelsdict:
            if "repoName" and "dataModels" in item:
                if item["repoName"] == subject:
                    output = item["repoLink"]
                    done = True
    if not done:
        output = False

    return output


# 15
def datamodel_repolink(datamodel: str):
    """It returns an array with the direct links to the repositories where is located the data model if it is found and False if not found.
    Parameters:
        datamodel: name of the data model

    Returns:
        if data model is found
            array of urls (even with one single result) to the github repository. Example for subject Activity it returns ['https://github.com/smart-data-models/dataModel.User.git']
        if data model is not found
            False
    """

    output = []
    done = False

    with open(official_list_file_name, "r", encoding='utf-8') as list_of_datamodels_pointer:
        datamodelsdict = json.load(list_of_datamodels_pointer)["officialList"]
        for item in datamodelsdict:
            if "repoName" and "dataModels" in item:
                dataModels = item["dataModels"]
                if datamodel in dataModels:
                    output.append(item["repoLink"])
                    done = True
    if not done:
        output = False

    return output


# 16
def update_data():
    """Download the latest data files from a remote server
    """

    data_dir = os.path.join(os.path.dirname(__file__), "model-assets")

    # Download the latest data files from a remote server
    urllib.request.urlretrieve(
        "https://raw.githubusercontent.com/smart-data-models/data-models/master/specs/AllSubjects/official_list_data_models.json",
        os.path.join(data_dir, "official_list_data_models.json"))
    urllib.request.urlretrieve("https://smartdatamodels.org/extra/smartdatamodels.json",
                               os.path.join(data_dir, "smartdatamodels.json"))
    urllib.request.urlretrieve("https://smartdatamodels.org/extra/datamodels_metadata.json",
                               os.path.join(data_dir, "datamodels_metadata.json"))

    # Update the data files with the latest information
    # (This will depend on the specific data files and the format they use)


# 17
def ngsi_ld_example_generator(schema_url: str):
    """It returns a fake normalized ngsi-ld format example based on the given json schema
    Parameters:
        schema_url: url of the schema (public available). (i.e. raw version of a github repo https://raw.githubusercontent.com/smart-data-models/dataModel.Aeronautics/master/AircraftModel/schema.json

    Returns:
        if the input parameter exists and the json schema is a valid json:
            a fake normalized ngsi-ld format example stored in dictionary format
        if there's any problem related to input parameter and json schema:
            False
    """

    dataModel = extract_datamodel_from_raw_url(schema_url)
    subject = extract_subject_from_raw_url(schema_url)

    # Both dataModel and subject need to be successfully extracted
    # dataModel is not in the list of pending that are unable to generate
    if dataModel == "dataModel" or subject == "subject": return False
    if dataModel in pending_datamodels: return returnPendingMessage

    payload = open_jsonref(schema_url)
    if payload == "": return False

    output = {}
    fullDict = {}

    # Parse the "allOf", "anyOf", "oneOf" structure
    if "allOf" in payload:
        for index in range(len(payload["allOf"])):
            if "properties" in payload["allOf"][index]:
                fullDict = {**fullDict, **payload["allOf"][index]["properties"]}
            else:
                fullDict = {**fullDict, **payload["allOf"][index]}
    elif "anyOf" in payload:
        for index in range(len(payload["anyOf"])):
            if "properties" in payload["anyOf"][index]:
                fullDict = {**fullDict, **payload["anyOf"][index]["properties"]}
            else:
                fullDict = {**fullDict, **payload["anyOf"][index]}
    elif "oneOf" in payload:
        for index in range(len(payload["oneOf"])):
            if "properties" in payload["oneOf"][index]:
                fullDict = {**fullDict, **payload["oneOf"][index]["properties"]}
            else:
                fullDict = {**fullDict, **payload["oneOf"][index]}
    else:
        fullDict = payload["properties"].copy()

    for prop in fullDict:

        parsedProperty = parse_property2ngsild_example({prop: fullDict[prop]}, dataModel, 0)

        # id and type should be key-value format in ngsild format
        if prop in ["id"]:
            output = {**output, **parsedProperty}
        elif prop in ["type"]:
            output = {**output, **{prop: parsedProperty}}
        else:
            output = {**output, **{prop: parsedProperty}}

    output["@context"] = [create_context(subject)]

    return output


# 18
def ngsi_ld_keyvalue_example_generator(schema_url: str):
    """It returns a fake key value ngsi-ld format example based on the given json schema
    Parameters:
        schema_url: url of the schema (public available). (i.e. raw version of a github repo https://raw.githubusercontent.com/smart-data-models/dataModel.Aeronautics/master/AircraftModel/schema.json

    Returns:
        if the input parameter exists and the json schema is a valide json:
            a fake key value ngsi-ld format example stored in dictionary format
        if there's any problem related to input parameter and json schema:
            False
    """

    output = ngsi_ld_example_generator(schema_url)
    if not output: return output
    if output == returnPendingMessage: return returnPendingMessage

    # normalized2keyvalues is deprecated
    keyvalues = normalized2keyvalues_v2(output)

    return keyvalues


# 19
def geojson_features_example_generator(schema_url: str):
    # TODO: Add ranges to the generated values of the different attributes
    """It returns a fake geojson feature format example based on the given json schema
    Parameters:
        schema_url: url of the schema (public available). (i.e. raw version of a GitHub repo https://raw.githubusercontent.com/smart-data-models/dataModel.Aeronautics/master/AircraftModel/schema.json

    Returns:
        if the input parameter exists and the json schema is a valide json:
            a fake geojson feature format example stored in dictionary format
        if there's any problem related to input parameter and json schema:
            False
    """

    noGeometryMessage = "\"wrong data model for generation of geojson Feature. No geographic properties\""
    output = ngsi_ld_example_generator(schema_url)
    if not output: return output
    if output == returnPendingMessage: return returnPendingMessage

    basePayload = output
    geopropertyName = ""
    # check that it is possible to generate the geojson feature
    if all(i in basePayload for i in ["id", "type", "location"]):
        geopropertyName = "location"
    else:
        for prop in basePayload:
            if isinstance(basePayload[prop], dict) and ('type' in basePayload[prop]):
                if basePayload[prop]["type"] == "GeoProperty":
                    geopropertyName = prop
        if geopropertyName == "":
            return "{\"error\": " + noGeometryMessage + "}"
    if geopropertyName != "":
        geojsonFeature = {}
        geojsonFeature["id"] = basePayload["id"]
        geojsonFeature["type"] = "Feature"
        geojsonFeature["geometry"] = basePayload[geopropertyName]["value"]
        rawPayload = basePayload.copy()
        rawPayload.pop("id")
        rawPayload.pop("type")
        geojsonFeature["properties"] = rawPayload
        return geojsonFeature


# 20
def update_broker(datamodel, subject, attribute, value, entityId=None, serverUrl=None, broker_folder="/ngsi-ld/v1",
                  updateThenCreate=True):
    # inspired by @antonio_jara
    """Update a broker compliant with a specific data model

    Parameters:
        - datamodel: the name of the data model of the SDM (see https://github.com/smart-data-models/data-models/blob/master/specs/AllSubjects/official_list_data_models.json) or this form
        https://smartdatamodels.org/index.php/list-of-data-models-3/
        - subject: the name of the subject, including the prefix dataModel.'subject'
        - attribute: name of the attribute in this data model. The list of available attributes can be found with this form
        https://smartdatamodels.org/index.php/ddbb-of-properties-descriptions/?wdt_column_filter%5B1%5D=WeatherObserved (change WeatherObserved by your data model) or with the function of the pysmartdatamodels sdm.attributes_datamodel(subject, datamodel)
        - value: the value to be updated or inserted
        - entity_id (str): The ID of the entity to update. If left none then the query for the broker is returned
        - serverUrl (str): The URL of the NGSI-LD broker.
        - broker_folder(str): It is supposed that the broker is installed in /ngsi/ld (default) change it if installed other location
        - updateThenCreate: If the updating attribute is nonexistent in the wanting entity, then create the attribute first if updateThenCreate is True, otherwise the operation is illegal

    Returns:
        - An array with two values
        - First the boolean result of the operation True if successful and False if not
        - Second a textual context explanation in every case
    """

    # from pysmartdatamodels import pysmartdatamodels as sdm

    def insert_in_broker(broker_url, entity):
        import requests
        import json

        # Convert the entity to JSON-LD format
        json_ld_entity = json.dumps(entity)

        try:
            # Send a POST request to the NSI-LD broker to insert the entity
            response = requests.post(broker_url + broker_folder + "/entities", data=json_ld_entity,
                                     headers={"Content-Type": "application/ld+json"})

            # Check the response status code
            if response.status_code == 201:
                return [True, "Successfully inserted in server " + broker_url + " the entity " + str(entity)]
            else:
                print("Failed to insert entity. Status code:", response.status_code)
                return [False, "Failed to insert entity in broker :" + broker_url + " with the status code" + str(
                    response.status_code) + " and this Response:" + response.text + " for this payload " + str(entity)]
        except requests.exceptions.RequestException as e:
            return [False, "An error  type (" + str(e) + ") occurred inserting the entity in broker :" + broker_url]

    def update_entity_in_broker(entity_id, updated_properties, broker_url):
        """Update an entity in an NGSI-LD broker.

        Parameters:
            - entity_id (str): The ID of the entity to update.
            - updated_properties (dict): A dictionary containing the updated properties of the entity.
            - broker_url (str): The URL of the NGSI-LD broker.

        Returns:
            - bool: True if the update was successful, False otherwise.
        """
        headers = {
            "Content-Type": "application/ld+json",
        }

        entity_url = f"{broker_url}{broker_folder}/entities/{entity_id}/attrs"
        print(entity_url)

        try:
            response = requests.patch(entity_url, json=updated_properties, headers=headers)

            if response.status_code == 204:  # 204 indicates a successful update with no content response
                return True
            elif response.status_code == 207:
                if updateThenCreate:
                    ngsi_type = ngsi_datatype_attribute(subject, datamodel, attribute)
                    create_properties = {attribute: {"type": ngsi_type, "value": value}}
                    create_headers = {
                        'Content-Type': 'application/json',
                        'Link': f'<{create_context(subject)}>; rel="http://www.w3.org/ns/json-ld#context"; type="application/ld+json"'
                    }

                    try:
                        create_response = requests.post(entity_url, json=create_properties, headers=create_headers)
                        if create_response.status_code == 204:  # 204 indicates a successful update with no content response
                            return True
                        else:
                            print(
                                f"Failed to update entity with nonexistent attribute. Status code: {create_response.status_code}.")
                            print(create_response.text)
                            return False

                    except requests.exceptions.RequestException as e:
                        print(f"Failed to update entity with nonexistent attribute: {e}")
                        return False
                else:
                    print(
                        f"Failed to update entity. Trying to update nonexistent attributes. If so, set updateThenCreate to true.")
                    return False
            else:
                print(f"Failed to update entity. Status code: {response.status_code}.")
                print(response.text)
                return False

        except requests.exceptions.RequestException as e:
            print(f"Failed to update entity: {e}")
            return False

    # load list of DM
    dataModelsListUrl = "https://raw.githubusercontent.com/smart-data-models/data-models/master/specs/AllSubjects/official_list_data_models.json"
    dataModelsList = open_jsonref(dataModelsListUrl)["officialList"]
    # for repo in dataModelsList:
    #     print(repo)
    dataModels = [repo["dataModels"] for repo in dataModelsList]
    subjects = [repo["repoName"] for repo in dataModelsList]
    # print(dataModels)
    # all list of data models (even repeated ones)
    full_list_datamodels = [element for sublist in dataModels for element in sublist]
    # check if data model exists
    if datamodel not in full_list_datamodels:
        return [False, "Data model: " + datamodel + " not found in any subject"]
    # check if subject exists
    if subject not in subjects:
        return [False, "Subject: " + subject + " not found. Have you include the full dataModel.xxxx name?"]
    # check if data model exists in this subject
    elif subject in subjects:
        found_data_model = False
        if datamodel not in datamodels_subject(subject):
            return [False, "the data model : " + datamodel + " is not found in the subject: " + subject]
        else:
            if attribute not in attributes_datamodel(subject, datamodel):
                return [False, "The attribute : " + attribute + " is not found in the data model: " + datamodel]
            else:
                data_type = datatype_attribute(subject, datamodel, attribute)
                # Define your JSON schema
                schema = {
                    "type": data_type
                }
                try:
                    # Validate the variable against the schema
                    validate(value, schema)
                    print("Variable matches the specified data type.")

                    if entityId is None:  # we need to create an entity
                        id_random = "urn:ngsi-ld" + generate_random_string(length=4) + ":" + generate_random_string(
                            length=4)
                        if serverUrl is not None:
                            ngsi_type = ngsi_datatype_attribute(subject, datamodel, attribute)
                            if ngsi_type == "Relationship":
                                payload = {"id": id_random, "type": datamodel,
                                           attribute: {"type": ngsi_type, "object": value},
                                           "@context": create_context(subject)}
                            else:
                                payload = {"id": id_random, "type": datamodel,
                                           attribute: {"type": ngsi_type, "value": value},
                                           "@context": create_context(subject)}
                            return insert_in_broker(serverUrl, payload)
                        else:  # no entity id and no broker, just returned the payload to be inserted
                            return [True, {"id": id_random, "type": datamodel, attribute: value,
                                           "@context": create_context(subject)}]

                    else:  # there is an entity so let's update
                        payload = {"id": entityId, "type": datamodel, attribute: value,
                                   "@context": create_context(subject)}
                        if serverUrl is not None:
                            result = update_entity_in_broker(entityId, payload, serverUrl)
                            if result:
                                return [result, "succesfully updated this payload" + str(payload)]
                            else:
                                return [result, "There is a problem updating this payload" + str(payload)]
                        else:
                            return [True, "the payload for the update should be this one:" + str(payload)]


                except jsonschema.exceptions.ValidationError as e:
                    print("Variable does not match the specified data type.")
                    print(e)
                    return [False, "The attribute: " + attribute + " cannot store the value : " + str(value)]


# 21
def generate_sql_schema(model_yaml: str) -> str:
    """
    Generate a PostgreSQL schema SQL script from the model.yaml representation of a Smart Data Model.

    Parameters:
        model_yaml (str): url of the model.yaml file (public available). (i.e. raw version of a github repo https://raw.githubusercontent.com/smart-data-models/dataModel.Weather/master/WeatherAlert/model.yaml)

    Returns:
        str: A string containing the PostgreSQL schema SQL script.
    """

    # open yaml
    model_yaml = open_yaml(model_yaml)

    # Get the entity name
    entity = list(model_yaml.keys())[0]

    # Initialize SQL schema statements
    sql_schema_statements = []
    sql_type_statement = []

    sql_data_types = ""

    # Define format mappings for YAML formats to postgreSQL Schema types
    type_mapping = {
        "string": "TEXT",
        "integer": "INTEGER",
        "number": "NUMERIC",
        "boolean": "BOOLEAN",
        "object": "JSON",
        "array": "JSON",
    }

    # Define format mappings for YAML formats to postgreSQL Schema types
    format_mapping = {
        "date-time": "TIMESTAMP",
        "date": "DATE",
        "time": "TIME",
        "uri": "TEXT",
        "email": "TEXT",
        "idn-email": "TEXT",
        "hostname": "TEXT",
        "duration": "TEXT"
    }

    # Start by creating the table
    table_create_statement = f"CREATE TABLE {entity} ("

    for key, value in model_yaml[entity]["properties"].items():
        field_type = "JSON"  # Default to JSON if type is not defined

        # Field type mapping
        if "type" in value:
            if "format" in value:
                # format type mapping (format overrides type)
                field_type = format_mapping.get(value["format"])
                # add attribute to the SQL schema statement
                sql_schema_statements.append(f"{key} {field_type}")

            elif "enum" in value:
                enum_values = value["enum"]
                enum_values = [str(element) for element in enum_values]
                if key == "type":
                    field_type = f"{entity}_type"
                else:
                    field_type = f"{key}_type"
                # create sql create type statment
                sql_type_statement.append(f"CREATE TYPE {field_type} AS ENUM ({','.join(map(repr, enum_values))});")

                sql_data_types += "CREATE TYPE " + field_type + " AS ENUM ("
                sql_data_types += f"{','.join(map(repr, enum_values))}"
                sql_data_types += ");"

                # add attribute to the SQL schema statement
                sql_schema_statements.append(f"{key} {field_type}")

            else:
                field_type = type_mapping.get(value["type"])
                # add attribute to the SQL schema statement
                sql_schema_statements.append(f"{key} {field_type}")
        elif "oneOf" in value:
            field_type = "JSON"
            sql_schema_statements.append(f"{key} {field_type}")

        # Handle the case when "allOf" exists
        if key == "allOf" and isinstance(value, list):
            for values in value:
                for sub_key, sub_value in values.items():
                    if isinstance(sub_value, dict):
                        if "format" in sub_value:
                            sub_field_type = format_mapping.get(sub_value["format"])
                            sql_schema_statements.append(f"{sub_key} {sub_field_type}")
                        if "type" in sub_value:
                            sub_field_type = type_mapping.get(sub_value["type"])
                            sql_schema_statements.append(f"{sub_key} {sub_field_type}")

        if key == "id":
            field_type = "TEXT PRIMARY KEY"
            # add attribute to the SQL schema statement
            sql_schema_statements.append(f"{key} {field_type}")

    # Complete the CREATE TABLE statement
    table_create_statement += ", ".join(sql_schema_statements)
    table_create_statement += ");"
    # PostgreSQL schema 
    result = sql_data_types + "\n" + table_create_statement
    print(result)

    return result


# 22
def look_for_datamodel(datamodel_search_text, approx_percentage=0):
    """Look for the data models that match the text included

    Parameters:
        - datamodel_search_text: piece of string with a text to be searched across the name of the data model of the SDM (It can be a part of the data model name)
        - approx_percentage: optional parameter, in case a non-exact matching is required, percentage of likelikhood between 0 and 99. i.e. 90 means quite similar. Example Weather is 64% similar to WeatherObserved, or 74% to WeatherAlert

    Returns:
        - An array with the values of the matching data models names

    """
    from fuzzywuzzy import fuzz
    # library to enable the fuzzy search

    datamodels = list_all_datamodels()
    output = []
    if approx_percentage == 0:
        output = [dm for dm in datamodels if datamodel_search_text in dm]
    else:
        output = [dm for dm in datamodels if (fuzz.ratio(dm, datamodel_search_text) > approx_percentage)]
    return output


# 23
def list_datamodel_metadata(datamodel, subject):
    """Look for the metadata data models that match the text included

       Parameters:
           - datamodel: Exact name of the data model
           - subject: Exact name of the subject

       Returns:
           - An object  with the metadata available for this data model in the different subjects
                -   version
                i.e. "0.0.2"
                -   modelTags
                i.e. "GreenMov"
                -   title of the data model
                i.e. "Smart Data Models. User Context schema"
                -   $id single link to the schema"
                i.e. https://smart-data-models.github.io/dataModel.User/UserContext/schema.json"
                -   description del data model
                i.e. "Information on the context of an anonymized in a given point in time"
                -   required attributes (array)
                i.e. ["id", "type"]
                -   yamlUrl link to the yaml version of the schema (model.yaml)
                i.e. https://raw.githubusercontent.com/smart-data-models/dataModel.User/master/UserContext/model.yaml
                -   jsonSchemaUrl link to the schema (schema.json)
                i.e. https://raw.githubusercontent.com/smart-data-models/dataModel.User/master/UserContext/schema.json"
                -   @context location of the @context of the subject.
                i.e. "https://raw.githubusercontent.com/smart-data-models/dataModel.User/master/context.jsonld"
                - exampleKeyvaluesJson link to the example in key values for NGSI v2 (json)
                i.e. "https://raw.githubusercontent.com/smart-data-models/dataModel.User/master/UserContext/examples/example.json"
                 - exampleKeyvaluesJsonld link to the example in key values for NGSI LD (jsonld)
                 i.e. "https://raw.githubusercontent.com/smart-data-models/dataModel.User/master/UserContext/examples/example.jsonld"
                 - exampleNormalizedJson link to the example in normalized for NGSI v2 (json)
                 i.e. "https://raw.githubusercontent.com/smart-data-models/dataModel.User/master/UserContext/examples/example-normalized.json"
                 - exampleNormalizedJsonld link to the example in normalized for NGSI LD (jsonld)
                 i.e. "https://raw.githubusercontent.com/smart-data-models/dataModel.User/master/UserContext/examples/example-normalized.jsonld"
                 - sql sql export of the schema.
                 i.e. "https://raw.githubusercontent.com/smart-data-models/dataModel.User/master/UserContext/schema.sql"
                 - adopters	list of the adopters of the data model
                 i.e. "https://raw.githubusercontent.com/smart-data-models/dataModel.User/master/UserContext/ADOPTERS.yaml"
                 - contributors	list of the contributors of the subject
                 i.e. "https://raw.githubusercontent.com/smart-data-models/dataModel.User/master/CONTRIBUTORS.yaml"
                 - spec Specification in English
                 i.e. "https://raw.githubusercontent.com/smart-data-models/dataModel.User/master/UserContext/doc/spec.md"
                 - spec_DE Specification in German
                 i.e. "https://raw.githubusercontent.com/smart-data-models/dataModel.User/master/UserContext/doc/spec_DE.md"
                 - spec_ES Specification in Spanish
                 i.e. "https://raw.githubusercontent.com/smart-data-models/dataModel.User/master/UserContext/doc/spec_ES.md"
                 - spec_FR Specification in French
                 i.e. "https://raw.githubusercontent.com/smart-data-models/dataModel.User/master/UserContext/doc/spec_FR.md"
                 - spec_IT Specification in Italian
                 i.e. "https://raw.githubusercontent.com/smart-data-models/dataModel.User/master/UserContext/doc/spec_IT.md"
                 - spec_JA Specification in Japanese
                 i.e. "https://raw.githubusercontent.com/smart-data-models/dataModel.User/master/UserContext/doc/spec_JA.md"
                 - spec_KO Specification in Korean
                 i.e. "https://raw.githubusercontent.com/smart-data-models/dataModel.User/master/UserContext/doc/spec_KO.md"
                 - spec_ZH Specification in Chinese
                 i.e. "https://raw.githubusercontent.com/smart-data-models/dataModel.User/master/UserContext/doc/spec_ZH.md"
           - if the data model is not found it returs False
           """

    output = []

    # Opens the file with the metadata about the data models
    with open(metadata_file, "r", encoding='utf-8') as metadata_file_pointer:
        metadata_dict = json.load(metadata_file_pointer)

    # It looks if this datamodel-subject pair is in the metadata. Although only one result is expected execution this way is quicker than other alternatives
    output = [metadata for metadata in metadata_dict if
              (metadata["dataModel"] == datamodel) and (metadata["subject"] == subject)]

    # returns the outcome
    if len(output) == 0:
        return False
    else:
        return output[0]

    # def validate_payload(datamodel, subject, payload):
    # TODO: Add this function to the package in next version
    """Validates a payload against a data model 

       Parameters:
           - datamodel: Exact name of the data model
           - subject: Exact name of the subject
           - payload: key values format of payload to be validated

       Returns:
           - An object with two subattributes 
                -   result: True or false
                -   details: Details about the successful validation 
                -   Case True: I.e. if there is a warning like exceeding attributes present but not in the data model)
                -   Case False: I.e. the outcome of the validation library

        Remarks for the development
        -   Maybe the libraries jsonschema, json-checker, pydantic could be useful here
        - Structure: 
        -   validate the inputs
        -       validate that the payload is a real and valid json
        -       validate that there is a type attribute coinciding to the data model name
        -       validate that the pointed data model really exist in the database 
    """

    # def create_QR_code(datamodel, subject):
    # TODO: Add this function to the package in next version
    """Look for the metadata data models that match the text included

       Parameters:
           - datamodel: Exact name of the data model
           - subject: Exact name of the subject

       Returns:
           - An image payload  
                

        Remarks for the development
        -   This code below creates such QR 
        - It has to 
        -   1) be adapted to the input parameters and to the output format
        - Remove own functions already present in the package
        -   2) include the new required packages in the requirements of the package
        - Maybe this internal procedure for extending the package can help on getting all the details https://smartdatamodels.org/index.php/procedure-to-extend-pysmartdatamodels-package/ of course not all tasks need to be accomplished if you are extending the package in your PR  
        
        def create_qr_image(filename, repo_name, dataModel):
    import qrcode
    from PIL import Image, ImageDraw, ImageFont
    from github import Github

    def open_json(fileUrl):
        import json
        import requests
        if fileUrl[0:4] == "http":
            # es URL
            try:
                pointer = requests.get(fileUrl)
                return json.loads(pointer.content.decode('utf-8'))
            except:
                return None

        else:
            # es file
            try:
                file = open(fileUrl, "r")
                return json.loads(file.read())
            except:
                return None
            # test

    logo_path = "smartdatamodels.png"
    credentialsFile = "/home/aabella/transparentia/CLIENTES/EU/FIWARE/credentials.json"
    # credentialsFile = "/home/fiware/credentials.json"
    credentials = open_json(credentialsFile)
    access_token = credentials["token"]
    repo_owner = credentials["globalUser"]
    g = Github(access_token)
    repo = g.get_organization(repo_owner).get_repo(repo_name)


    # Generate QR code image
    qr = qrcode.QRCode(
        version=None,
        error_correction=qrcode.constants.ERROR_CORRECT_L,
        box_size=10,
        border=3,
    )
    url = f"https://github.com/{repo_owner}/{repo_name}/blob/master/{dataModel}/doc/spec.md"

    qr.add_data(url)
    qr.make(fit=True)
    # img = qr.make_image(fill_color="#002e67", back_color="white")
    img = qr.make_image(fill_color="#002e67", back_color="white")
    border_size = 24
    border = Image.new("RGB", (img.size[0] + border_size * 2, img.size[1] + border_size * 2), "white")
    border.paste(img, (border_size, border_size))

    # Add the letter
    draw = ImageDraw.Draw(border)
    font = ImageFont.truetype("arial.ttf", 20)
    draw.text((border_size, border_size), "Structure of the Data model:" + dataModel, font=font, fill="#002e67")
    moreinfo  = "FIWARE Foundation. info@smartdatamodels.org"
    draw.text((border_size, border_size + img.size[0]), moreinfo, fill="#002e67", font=font,)
    # Add the logo
    logo = Image.open(logo_path)
    logo_size = int(border_size * 2.5)
    logo = logo.resize((logo_size, logo_size))
    border.paste(logo, (
    int((border_size * 2 + img.size[0] - logo_size) / 2), int((border_size * 2 + img.size[1] - logo_size) / 2)))

    # Save image to file
    border.save(filename)

    # Upload file to GitHub repository
    with open(filename, "rb") as f:
        content = f.read()
    # repo.create_file(dataModel + "/" + filename, "QR image", content)
 
    """

    # def include_local_datamodel(schema, subject, datamodel, contributors (optional), adopters (optional), notes(optional)):
    # TODO: Add this function to the package in next version
    """The aim of this function is to allow the users to include new data models locally 
    This information will not be shared with the central repository unless the user is really willing to do it with an additional functions (see submit datamodel function)

       Parameters:
           - schema: In json schema with the description meeting the contribution manual https://bit.ly/contribution_manual
           - subject: Exact name of the proposed local subject
           - datamodel: name of the proposed data model 
           - contributors: Optional parameter. List of the contributors of the data model (usually it is a pointer to an external file in yaml like this one https://github.com/smart-data-models/dataModel.Transportation/blob/master/CONTRIBUTORS.yaml
           - adopters: Optional parameter. List of the adopters of the data model (usually it is a pointer to an external file in yaml like this one https://github.com/smart-data-models/dataModel.Transportation/blob/master/APDSObservation/ADOPTERS.yaml
           
       Returns:
           - Extends the files located ad model-assets/smartdatamodels.json, official_list_data_models and datamodels_metadata.json.
           - As a result local data models will be treated absolutely like the official ones
                   
        Remarks for the development
        -   The new subject-datamodel pair should be different from an existing one to prevent overwriting existing data models
        - Maybe it could be useful for the development the current script we have for running the inventory of the database https://github.com/smart-data-models/data-models/blob/master/utils/30_f_properties_inventory_10.0.py
          
    """

    # def submit_datamodel(subject, datamodel, contributors (optional), adopters (optional), notes(optional), example_payload, notes_context, public_repository, credentials):
    # TODO: Add this function to the package in next version
    """The aim of this function is to allow the users to share their local data models to the central repository in order to get it accepted
    The function will create the structure expected in the local repository provided, eventually this local repository could be the incubated.  
    
       Parameters:
           - subject: Exact name of the proposed local subject
           - datamodel: name of the proposed data model 
           - contributors: Optional parameter. List of the contributors of the data model (usually it is a pointer to an external file in yaml like this one https://github.com/smart-data-models/dataModel.Transportation/blob/master/CONTRIBUTORS.yaml
           - adopters: Optional parameter. List of the adopters of the data model (usually it is a pointer to an external file in yaml like this one https://github.com/smart-data-models/dataModel.Transportation/blob/master/APDSObservation/ADOPTERS.yaml

       Returns:
           - A PR in the selected repository creating according to the structure described in the contribution manual https://docs.google.com/presentation/d/e/2PACX-1vTs-Ng5dIAwkg91oTTUdt8ua7woBXhPnwavZ0FxgR8BsAI_Ek3C5q97Nd94HS8KhP-r_quD4H0fgyt3/pub?start=false&loop=false&delayms=3000#slide=id.g280c21c8f96_0_124 (page 10 contribution manual)
           - Eventually a mail to info@smartdatamodels.org announcing the availability of the repo in order to start the accepting process.

        

    """


# 24
def validate_dcat_ap_distribution_sdm(json_data):
    """The function takes a distribution DCAT-AP in json and validates if the downloadURL contains a valid payload against conformsTo.
    It works when the downloadURL contains a json keyvalues payload and
    conformsTo point to the raw version of a Smart Data Model
       Parameters:
           json_data: metadata in DCAT-AP distribution format (See https://github.com/smart-data-models/dataModel.DCAT-AP/blob/master/Distribution/schema.json and
           https://semiceu.github.io/DCAT-AP/releases/3.0.0/
                conformsTo: Attribute (Array) with links to the json schemas of SDM
                downloadURL: Public available link to the raw format of the payload

       Returns:
           It prints a version of the attributes separated by the separator listing the meta_attributes specified
           A variable with the same strings
           if any of the input parameters is not found it returns False
    """

    validated = False
    try:
        conforms_to = json_data.get('conformsTo', [])  # retrieves the attribute conformsTo
    except:
        return False

    ################## SECTION FOR VALIDATION WITH SMART DATA MODELS ##################
    valid_link = any(link.startswith('https://github.com/smart-data-models') or link.startswith(
        "https://raw.githubusercontent.com/smart-data-models") for link in conforms_to)
    # it checks that the link belong to smart data Models
    # it should point to the raw version of the json schema
    if not valid_link:
        print(
            'Warning: "conformsTo" attribute should contain at least one link belonging to the domain github.com/smart-data-models')
        return validated
    schemas = [link for link in conforms_to if
               link.startswith('https://github.com/smart-data-models') or link.startswith(
                   "https://raw.githubusercontent.com/smart-data-models")]
    # only those schemas belonging to Smart Data Models are analysed

    # Check if 'downloadURL' attribute points to a valid location to download a file
    if json_data.get('downloadURL', '') is None:
        print('Warning: "I cannot access the distribution URL')
        return validated
    else:
        download_url = json_data.get('downloadURL')
        print(download_url)

    for distribution in download_url:
        payload = open_jsonref(distribution)  # It retrieves the content of the payload to be validated
        for schema in schemas:
            print(schema)
            schema = open_jsonref(schema)  # It retrieves the schema for validating the payload
            try:
                validate(instance=payload, schema=schema)
                print("The download URL content is validated by the JSON schema.")
                # If you reach this point, the content is validated by the current schema
                validated = True
                break
            except Exception as e:
                # If the content is not validated by the current schema, try the next one
                continue
        else:
            print("The download URL content is not validated by any of the JSON schemas.")
            return validated

    return validated
    ################## END OF SECTION FOR VALIDATION WITH SMART DATA MODELS ##################


# 25
def subject_for_datamodel(datamodel):
    """The function looks for the subject corresponding to a data model name
    if not found it returns false
       Parameters:
           datamodel: name of the data model

       Returns:
           An array (always) if there is only one element with the names of the subjects
           Usually only one element in the array is returned because there are few clashes in data model names
           False if no subject is found
    """

    with open(official_list_file_name, "r", encoding='utf-8') as list_of_datamodels_pointer:
        list_of_datamodels = json.load(list_of_datamodels_pointer)["officialList"]
        subjects = [repo["repoName"] for repo in list_of_datamodels if datamodel in repo["dataModels"]]
    if len(subjects) == 0:
        return False
    else:
        return subjects


# 26
def validate_payload(datamodel, subject, payload):
    """Look for the metadata data models that match the text included

       Parameters:
           - datamodel: Exact name of the data model
           - subject: Exact name of the subject
           - payload: key values format of payload to be validated

       Returns:
           - An object with two subattributes
                -   result: True or false
                -   details: Details about the succesful validation
                -   Case True: I.e. if there is a warning like exceeding attributes present but not in the data model)
                -   Case False: I.e. the outcome of the validation library

        Remarks for the development
        -   Maybe the libraries jsonschema, json-checker, pydantic could be useful here
        - Structure:
        -   validate the inputs
        -       validate that the payload is a real and valid json
        -       validate that there is a type attribute coinciding to the data model name
        -       validate that the pointed data model really exist in the database
    """
    output = [False, "No validation performed"]
    subjects = load_all_datamodels()
    datamodels_found = [datamodel for subject in subjects if datamodel in subject["dataModels"]]
    if datamodels_found:
        return [True, datamodels_found]
    else:
        return [False, "datamodel/subject not found"]
# result = validate_payload("WeatherObserved", "dataModel.Weather", "")
# print(result)
